## Brian Blaylock
## July 29, 2021

"""
Some initial work to use Herbie to access HRRR-Zarr

The directory structure is very different than the GRIB2 format.

- https://hrrrzarr.s3.amazonaws.com/index.html
- https://mesowest.utah.edu/html/hrrr/

"""

import xarray as xr
import s3fs
import pandas as pd

# Let's grab an analysis file using s3fs

fs = s3fs.S3FileSystem(anon=True)

date = pd.to_datetime("2021-07-03 12:00")
level = "2m_above_ground"
var = "TMP"

url = f"hrrrzarr/sfc/{date:%Y%m%d}/{date:%Y%m%d_%H}z_anl.zarr/{level}/{var}/{level}/"
store = s3fs.S3Map(root=url, s3=fs, check=False)
ds = xr.open_zarr(store)

ds.TMP.plot()



#!/usr/bin/env python3

## Brian Blaylock
## July 8, 2020

"""
👴🏻 The HRRR-B API is deprecated. Use the new Herbie API.

=======================================
Retrieve HRRR GRIB2 files from archives
=======================================

Download High-Resolution Rapid Refresh (HRRR) model GRIB2 files from 
different archive sources. Support for subsetting GRIB2 files by fields
if the `.idx`` file exist.

Default HRRR Data Source Priority
---------------------------------
NOAA started pushing the HRRR data to the Google Cloud Platform and 
Amazon Web Services at the end of 2020. The archives were also 
backfilled to previous years as far back as July 30, 2014. This module
will download HRRR data from these sources in the following order (the
default download source priority can be changed).


1. NOMAS: NOAA Operational Model Archive and Distribution System
    - https://nomads.ncep.noaa.gov/
    - Available for today's and yesterday's runs
    - Original data source. All available data.
    - Includes ``.idx`` for all GRIB2 files.

2. Google: Google Cloud Platform Earth
    - https://console.cloud.google.com/storage/browser/high-resolution-rapid-refresh
    - Available from July 30, 2014 to Present.
    - Does not have ``.idx`` files before September 17, 2018.
    - Has all original data including nat, subh, prs, and sfc files 
      for all forecast hours.

3. AWS: Amazon Web Services
    - https://noaa-hrrr-bdp-pds.s3.amazonaws.com/
    - Available from July 30, 2014 to Present.
    - Does not have ``.idx`` files.
    - Has all nat, subh, prs, and sfc files for all forecast hours.
    - Some data may be missing.

3. Pando: The University of Utah HRRR archive
    - http://hrrr.chpc.utah.edu/
    - Available from July 15, 2016 to Present.
    - A subset of prs and sfc files.
    - Contains a ``.idx`` file for *every* GRIB2 file for subsetting.

Main Functions
--------------
download_hrrr
    Download GRIB2 files to local disk.
xhrrr
    Download and read HRRR data as an xarray.Dataset with cfgrib engine.

"""

import os
import re
from datetime import datetime, timedelta
from pathlib import Path
from itertools import product
import warnings
import configparser

import urllib.request
import requests
import cfgrib
import numpy as np
import pandas as pd
import xarray as xr

from hrrrb.tools import to_180, get_crs

warnings.warn(
      "The hrrrb API is deprecated. Use the new Herbie API instead."
   )

#=======================================================================
# Specify default location to save HRRR GRIB2 files
#=======================================================================
config = configparser.ConfigParser()
_config_path = Path('~').expanduser() / '.config' / 'hrrrb' / 'config.cfg'

user_home_default = str(Path('~').expanduser() / 'data')

if not _config_path.exists():
    _config_path.parent.mkdir(parents=True)
    _config_path.touch()
    config.read(_config_path)
    config.add_section('download')
    config.set('download', 'default_save_dir', user_home_default)
    with open(_config_path, 'w') as configfile:
        config.write(configfile)
    print(f'⚙ Created config file [{_config_path}]',
          f'with default download directory set as [{user_home_default}]')

config.read(_config_path)
try:
    _default_save_dir = Path(config.get('download', 'default_save_dir'))
except:
    print(f'🦁🐯🐻 oh my! {_config_path} looks weird,',
          f'but I will add a new section')
    config.add_section('download')
    config.set('download', 'default_save_dir', user_home_default)
    with open(_config_path, 'w') as configfile:
        config.write(configfile)
    _default_save_dir = Path(config.get('download', 'default_save_dir'))

# Uncomment to just set save dir to the user's home directory
#_default_save_dir = Path('~').expanduser() / 'data'
#=======================================================================
#=======================================================================


# List of HRRR download source base URLs, in order of priority
base_url = dict(
    nomads = 'https://nomads.ncep.noaa.gov/pub/data/nccf/com/hrrr/prod',
    google = 'https://storage.googleapis.com/high-resolution-rapid-refresh',
    aws = 'https://noaa-hrrr-bdp-pds.s3.amazonaws.com',
    pando = 'https://pando-rgw01.chpc.utah.edu',
    pando2 = 'https://pando-rgw02.chpc.utah.edu',  # The 2nd rados gateway might work if the 1st didn't.
)

def _ping_pando(gateway=1):
    """Pinging the Pando server before downloading can prevent a bad handshake."""
    try:
        requests.head(f'https://pando-rgw0{gateway}.chpc.utah.edu/')
    except:
        print('bad handshake...am I able to move on?')
        pass

def _reporthook(a, b, c):
    """
    Report download progress in megabytes. Prints progress to screen.

    Parameters
    ----------
    a : Chunk number
    b : Maximum chunk size
    c : Total size of the download
    """
    chunk_progress = a * b / c * 100
    total_size_MB =  c / 1000000.
    print(f"\r🚛💨  Download Progress: {chunk_progress:.2f}% of {total_size_MB:.1f} MB\r", end='')

def _searchString_help(searchString):
    msg = [
        f"There is something wrong with [[  searchString='{searchString}'  ]]",
        "\nUse regular expression to search for lines in the .idx file",
        "Here are some examples you can use for `searchString`",
        "  ============================= ===============================================",
        "  ``searchString``              Messages that will be downloaded",
        "  ============================= ===============================================",
        "  ':TMP:2 m'                    Temperature at 2 m.",
        "  ':TMP:'                       Temperature fields at all levels.",
        "  ':UGRD:.* mb'                 U Wind at all pressure levels.",
        "  ':500 mb:'                    All variables on the 500 mb level.",
        "  ':APCP:'                      All accumulated precipitation fields.",
        "  ':APCP:surface:0-[1-9]*'      Accumulated precip since initialization time",
        "  ':APCP:surface:[1-9]*-[1-9]*' Accumulated precip over last hour",
        "  ':UGRD:10 m'                  U wind component at 10 meters.",
        "  ':(U|V)GRD:(10|80) m'         U and V wind component at 10 and 80 m.",
        "  ':(U|V)GRD:'                  U and V wind component at all levels.",
        "  ':.GRD:'                      (Same as above)",
        "  ':(TMP|DPT):'                 Temperature and Dew Point for all levels .",
        "  ':(TMP|DPT|RH):'              TMP, DPT, and Relative Humidity for all levels.",
        "  ':REFC:'                      Composite Reflectivity",
        "  ':surface:'                   All variables at the surface.",
        "  ============================= ===============================================",
        "\nIf you need help with regular expression, search the web",
        "  or look at this cheatsheet: https://www.petefreitag.com/cheatsheets/regex/."
        ]
    return '\n'.join(msg)

def _outFile_from_url(url, save_dir):
    """
    Define the filename for a grib2 file from its URL.

    Parameters
    ----------
    url : str
        A URL string for a grib2 file we want to download.
    save_dir : dir
        The directory path to save the file.
    """
    if 'pando' in url:
        outFile = '_'.join(url.split('/')[-2:])
    else:
        outFile = url.split('/')[-3][5:] + '_' + url.split('/')[-1]

    outFile = save_dir / outFile

    if not outFile.parent.is_dir():
        outFile.parent.mkdir(parents=True, exist_ok=True)
        print(f'👨🏻‍🏭 Created directory: [{outFile.parent}]')

    return outFile

def _download_hrrr_subset(url, searchString, *,
                          df_row_urls=None,
                          save_dir=_default_save_dir,
                          dryrun=False, verbose=True):
    """
    Download a subset of GRIB fields from a HRRR file located at a URL.

    There must be an index (.idx) file available for the file.

    A USER, YOU SHOULD NOT USE THIS FUNCTION DIRECTLY. Instead, use the
    ``download_hrrr`` function and specify the search string to get a 
    subset of a GRIB2 file.

    Parameters
    ----------
    url : string
        The URL for the HRRR file you are trying to download. There must be an
        index file for the GRIB2 file. For example, if
        ``url='https://pando-rgw01.chpc.utah.edu/hrrr/sfc/20200624/hrrr.t01z.wrfsfcf17.grib2'``,
        then ``https://pando-rgw01.chpc.utah.edu/hrrr/sfc/20200624/hrrr.t01z.wrfsfcf17.grib2.idx``
        must also exist.
    searchString : str
        The string you are looking for in each line of the index file.
        Refer to the _searchString_help for examples.
    df_row_urls : None or list
        THIS VARIABLE IS RESERVED FOR INPUT FROM THE ``download_hrrr``
        FUNCTION for the case when there isn't an .idx file from the 
        requested source, the .idx might be available at a different URL.
        This is a list of those other URLs (supplied by the download_hrrr function's loop)
    save_dir : string
        Directory path to save the file, default ``~/data/``.
    dryrun : bool
        If True, do not perform download, but print out what the function will
        attempt to do.
    verbose : bool
        If True, print lots of details (default)

    Returns
    -------
    The path and name of the new file.
    """
    # Ping Pando first. This *might* prevent a "bad handshake" error.
    if 'pando-rgw01' in url: _ping_pando(1)
    elif 'pando-rgw02' in url: _ping_pando(2)

    # Make a request for the .idx file for the above URL
    idx = url + '.idx'
    r = requests.get(idx)

    # Check that the file exists. If there isn't an index, you will get a 404 error.
    if not r.ok:
        print(f'❌ .idx file does not exist at {idx}')
        if df_row_urls is not None:
            # Check the other sources provided
            for alt_url in df_row_urls[1:]:
                if alt_url != url:  # don't check the url twice
                    idx = alt_url + '.idx'
                    r = requests.get(idx)
                    if r.ok:
                        print(f'✔ I found a .idx file at {idx}')
                        continue
                    else:
                        print(f'❌ .idx file does not exist at {idx}')
    
    if not r.ok:
        notice = [
            '',
            f'❌ SORRY! Status Code: {r.status_code} {r.reason}',
            f'❌ It does not look like an index file exists: {idx}',
            f'❌ Please change your request or download the full grib2 file with `download_hrrr`.'
        ]
        raise Exception('\n'.join(notice))

    # Read the text lines of the request
    lines = r.text.split('\n')

    # Search expression
    try:
        expr = re.compile(searchString)
    except Exception as e:
        print('re.compile error:', e)
        raise Exception(_searchString_help(searchString))

    # Store the byte ranges in a dictionary
    #     {byte-range-as-string: line}
    byte_ranges = {}
    for n, line in enumerate(lines, start=1):
        # n is the line number (starting from 1) so that when we call for
        # `lines[n]` it will give us the next line. (Clear as mud??)

        # Use the compiled regular expression to search the line
        if expr.search(line):
            # aka, if the line contains the string we are looking for...

            # Get the beginning byte in the line we found
            parts = line.split(':')
            rangestart = int(parts[1])

            # Get the beginning byte in the next line...
            if n+1 < len(lines):
                # ...if there is a next line
                parts = lines[n].split(':')
                rangeend = int(parts[1])
            else:
                # ...if there isn't a next line, then go to the end of the file.
                rangeend = ''

            # Store the byte-range string in our dictionary,
            # and keep the line information too so we can refer back to it.
            byte_ranges[f'{rangestart}-{rangeend}'] = line

    if len(byte_ranges) == 0:
        # Loop didn't find the searchString in the index file.
        print(f'❌ WARNING: Sorry, I did not find [{searchString}] in the index file {idx}')
        print(_searchString_help(searchString))
        return None

    # What should we name the file we save this data to?
    # Let's name it something like `subset_20200624_hrrr.t01z.wrfsfcf17.grib2`
    runDate = list(byte_ranges.items())[0][1].split(':')[2][2:-2]
    outFile = '_'.join(['subset', runDate, url.split('/')[-1]])
    outFile = save_dir / outFile

    for i, (byteRange, line) in enumerate(byte_ranges.items()):

        if i == 0:
            # If we are working on the first item, overwrite the existing file.
            curl = f'curl -s --range {byteRange} {url} > {outFile}'
        else:
            # If we are working on not the first item, append the existing file.
            curl = f'curl -s --range {byteRange} {url} >> {outFile}'

        num, byte, date, var, level, forecast, *_ = line.split(':')

        if dryrun:
            if verbose: print(f'    🐫 Dry Run: Found GRIB line [{num:>3}]: variable={var}, level={level}, forecast={forecast}')
            #print(f'    🐫 Dry Run: `{curl}`')
        else:
            if verbose: print(f'  Downloading GRIB line [{num:>3}]: variable={var}, level={level}, forecast={forecast}')
            os.system(curl)

    if dryrun:
        if verbose: print(f'🌵 Dry Run: Success! Searched for [{searchString}] and found [{len(byte_ranges)}] GRIB fields. Would save as {outFile}')
    else:
        if verbose: print(f'✅ Success! Searched for [{searchString}] and got [{len(byte_ranges)}] GRIB fields and saved as {outFile}')

        return outFile

###############################################################################
###############################################################################

def download_hrrr(DATES, searchString=None, *, 
                  fxx=range(0, 1),
                  model='hrrr',
                  field='sfc',
                  save_dir=_default_save_dir,
                  download_source_priority=None,
                  overwrite=False, 
                  dryrun=False, verbose=True):
    """
    Download full or partial HRRR grib2 files for a list of dates and forecasts.

    Attempts to download grib2 files from three different sources in the
    following order:

    1. NOAA Operational Model Archive and Distribution System ('nomads').
    2. Google Cloud Platform ('google')
    3. University of Utah HRRR archive ('pando')
    4. Amazon Web Services ('aws')

    Parameters
    ----------
    DATES : datetime or list of datetimes
        A datetime or list of datetimes that represent the model
        initialization time for which you want to download. May also
        use a date as a string if it can be parsed by Pandas.
    searchString : str
        A regular expression string that describes the variables you 
        want to download from the file. This is used as the 
        `searchString` in ``_download_hrrr_subset`` to looking for 
        specific byte ranges from the file to download.

        Default is None, meaning to not search for all variables and 
        **downloads the full GRIB2 file**. Furthermore, ``':'`` is an 
        alias for None because it is equivalent to identifying every 
        line in the .idx file.

        Take a look at the .idx file at
        https://pando-rgw01.chpc.utah.edu/hrrr/sfc/20200624/hrrr.t01z.wrfsfcf17.grib2.idx
        for an example of what an index file looks like.

        Also look at this webpage: http://hrrr.chpc.utah.edu/HRRR_archive/hrrr_sfc_table_f00-f01.html
        for additional details. **You should focus your searches on the
        *variable* and *level* to key in on specific fields**.

        You may use regular expression syntax to customize your search.
        Check out this regular expression cheatsheet:
        https://link.medium.com/7rxduD2e06

        Here are a few examples that can help you get started

        ============================= ===============================================
        ``searchString``              Messages that will be downloaded
        ============================= ===============================================
        ':TMP:2 m'                    Temperature at 2 m.
        ':TMP:'                       Temperature fields at all levels.
        ':UGRD:.* mb'                 U Wind at all pressure levels.
        ':500 mb:'                    All variables on the 500 mb level.
        ':APCP:'                      All accumulated precipitation fields.
        ':APCP:surface:0-[1-9]*'      Accumulated precip since initialization time
        ':APCP:surface:[1-9]*-[1-9]*' Accumulated precip over last hour
        ':UGRD:10 m'                  U wind component at 10 meters.
        ':(U|V)GRD:'                  U and V wind component at all levels.
        ':.GRD:'                      (Same as above)
        ':(TMP|DPT):'                 Temperature and Dew Point for all levels .
        ':(TMP|DPT|RH):'              TMP, DPT, and Relative Humidity for all levels.
        ':REFC:'                      Composite Reflectivity
        ':surface:'                   All variables at the surface.
        ''             
        ============================= ===============================================

    fxx : int or list of int
        Forecast lead time or list of forecast lead times to download.
        Default only grabs analysis hour (f00). If you want all forecasts
        hours, you can set ``fxx=range(0,19)``.
    model : {'hrrr', 'hrrrak', 'hrrrX'}
        The model type you want to download.
        - `'hrrr'` HRRR Contiguous United States (operational)
        - `'hrrrak'` HRRR Alaska. You can also use 'alaska' as an alias.
        - `'hrrrX'` *experimental* HRRR (experimental version run at ESRL)
           is only available from the Pando archive.
        - `'rap'` Rapid Refresh model. **This is not fully tested or implemented**,
           but you can get the GRIB2 grids and subset the file by .idx with
           searchString. `field` is not required as all levels are in the same file.
    field : {'prs', 'sfc', 'nat', 'subh'}
        Variable fields you wish to download. Only 'sfc' and 'prs' are
        available on Pando, but 'nat' and 'subh' are attainable from
        other sources.
        - 'sfc' surface fields
        - 'prs' pressure fields
        - 'nat' native fields
        - 'subh' subhourly fields
    save_dir : pathlib.Path
        Directory path to save the downloaded HRRR files.
    download_source_priority : None or list
        If None, the default order is the order specified by
        ``base_url``. You may, however, specify a different
        priority by supplying a list of keys in the ``base_url``.
        For example, to download from 'google' before checking 'pando',
        you may do ``download_source_priority=['google', 'pando', 'nomads']``.
        This also makes is possible to exclude a source.
    overwrite : bool
        Only applicable if ``searchString=None``. Will check if the
        file exists, and if ``overwrite=False``, will not redownload 
        the file.
    dryrun : bool
        If True, instead of downloading the files, it will print out the
        files that would be downloaded. This is set to False by default.
    verbose :bool
        If True, print lots of information (default).
        If False, only print some info about download progress.

    Returns
    -------
    The file name for the HRRR files we downloaded and the URL it was from.
    For example: ``('20170101_hrrr.t00z.wrfsfcf00.grib2', 'https://pando-rgw01...')``
    """

    #*******************************************************************
    ## Check function input
    #*******************************************************************
       
    # The user may set `model='alaska'` as an alias for 'hrrrak'.
    if model.lower() == 'alaska': model = 'hrrrak'

    # Force the `field` input string to be lower case.
    field = field.lower()

    # `DATES` and `fxx` should be a list-like object, but if they have no
    # length, (like if the user requests a single date or forecast hour),
    # then turn the item into a list-like object.
    if not hasattr(fxx, '__len__'): fxx = [fxx]
    if isinstance(download_source_priority, str): download_source_priority = [download_source_priority]
    if isinstance(DATES, str): DATES = pd.to_datetime(DATES)
    if not hasattr(DATES, '__len__'): DATES = np.array([DATES])
    DATES = pd.to_datetime(DATES)

    if not all([i < datetime.utcnow() for i in DATES]):
        warnings.warn("🦨 Whoops! One or more of your DATES is in the future.")

    # Make save_dir if path doesn't exist
    if not hasattr(save_dir, 'exists'): 
        save_dir = Path(save_dir).resolve()

    save_dir = save_dir / model
    if not save_dir.is_dir():
        save_dir.mkdir(parents=True, exist_ok=True)
        print(f'👨🏻‍🏭 Created directory: [{save_dir}]')

    if download_source_priority is None:
        source_priority = base_url
    else:
        # If desired, reorder download_source_priority
        source_priority = {}
        for i in download_source_priority:
            i = i.lower()
            if i in base_url: 
                source_priority[i] = base_url[i]
            else: 
                print(f'  ♟ Skipping `{i}` as a potential source '
                      f'because it is not a valid source name. Must be '
                      f'one of {list(base_url)}.')

    # Ping Pando first. This *might* prevent a "bad handshake" error
    # when a file downloaded from Pando.
    if 'pando' in source_priority: _ping_pando(1)
    elif 'pando2' in source_priority: _ping_pando(2)

    #**************************************************************************
    # Build the URL path for every file we want
    #**************************************************************************
    # HRRR Full URL Examples
    # ----------------------
    # NOMADS
    #   https://nomads.ncep.noaa.gov/pub/data/nccf/com/hrrr/prod/hrrr.20200624/conus/hrrr.t00z.wrfsfcf09.grib2
    # Pando
    #   https://pando-rgw01.chpc.utah.edu/hrrr/sfc/20200624/hrrr.t01z.wrfsfcf17.grib2
    # Google
    #   https://storage.googleapis.com/high-resolution-rapid-refresh/hrrr.20200101/conus/hrrr.t00z.wrfnatf04.grib2
    # Amazon AWS
    #   https://noaa-hrrr-bdp-pds.s3.amazonaws.com/hrrr.20140917/conus/hrrr.t00z.wrfsubhf1015.grib2
    #
    # RAP Full URL Examples
    # ----------------------
    # NOMADS
    #   https://nomads.ncep.noaa.gov/pub/data/nccf/com/rap/prod/rap.20210407/rap.t00z.awip32f00.grib2
    # Amazon AWS
    #   https://noaa-rap-pds.s3.amazonaws.com/rap.20210407/rap.t00z.awip32f03.grib2
    # Google
    #   https://storage.googleapis.com/rapid-refresh/rap.20210222/rap.t00z.awip32f01.grib2

    # Create a list of URLs for each files from each download source
    URL_list = {}
    for source, source_url in source_priority.items():
        URL_list[source] = []
        
        for DATE, f in product(DATES, fxx):
            if source.startswith('pando'):
                # Pando URL is unique
                URL_list[source].append(f"{source_url}/{model}/{field}/{DATE:%Y%m%d}/{model}.t{DATE:%H}z.wrf{field}f{f:02d}.grib2")
            else:
                # NOMADS, Google, and AWS URL are the same after their base_url
                if model == 'hrrr':
                    URL_list[source].append(f"{source_url}/hrrr.{DATE:%Y%m%d}/conus/hrrr.t{DATE:%H}z.wrf{field}f{f:02d}.grib2")
                elif model == 'hrrrak':
                    URL_list[source].append(f"{source_url}/hrrr.{DATE:%Y%m%d}/alaska/hrrr.t{DATE:%H}z.wrf{field}f{f:02d}.ak.grib2")
                elif model == 'rap':
                    ## Rapid Refresh Model is a special case
                    if source == 'aws':
                        URL_list[source].append(f'https://noaa-rap-pds.s3.amazonaws.com/rap.{DATE:%Y%m%d}/rap.t{DATE:%H}z.awip32f{f:02d}.grib2')
                    elif source == 'nomads':
                        URL_list[source].append(f'https://nomads.ncep.noaa.gov/pub/data/nccf/com/rap/prod/rap.{DATE:%Y%m%d}/rap.t{DATE:%H}z.awip32f{f:02d}.grib2')
                    elif source == 'google':
                        URL_list[source].append(f'https://storage.googleapis.com/rapid-refresh/rap.{DATE:%Y%m%d}/rap.t{DATE:%H}z.awip32f{f:02d}.grib2')
                else:
                    URL_list[source].append(f'https://pando-rgw01.chpc.utah.edu/hrrrX/nonehere_placeholder_{source}') # Experimental HRRR (hrrrX) only exists on Pando. This is a placeholder.

    # A pandas.DataFrame to keep track of the URLs. We will iterate over each row.
    URL_df = pd.DataFrame(URL_list)

    #return URL_df

    #*******************************************************************
    # Download Files that exist
    #*******************************************************************
    # Attempt download from sources in order of source priority.

    n = len(URL_df)

    if dryrun:
        print(f'🌵 Info: Dry Run [{n}] GRIB2 files\n')
    else:
        print(f'💡 Info: Downloading [{n}] GRIB2 files\n')

    # Keep track of total time spent in the download loop.
    loop_time = timedelta()

    # We want to return lists of the filenames we retrive and the URLs
    all_files = []
    all_urls = []

    for (index, row), DATE in zip(URL_df.iterrows(), DATES):
        #print(index, row, DATE)
        #---------------------------------------------------------
        # Time keeping: *crude* method to estimate remaining time.
        #---------------------------------------------------------
        timer = datetime.now()
        #---------------------------------------------------------

        # We will set this to True when we have downloaded a file from a source
        # For example, this will prevents us from downloading a file from
        # google if we already got the file from pando.
        one_exists = False

        for source, url in row.items():
            # Check if the URL exists and the filesize is large enough to be a file.
            head = requests.head(url)
            check_exists = head.ok
            check_content = int(head.raw.info()['Content-Length']) > 1_000_000

            if check_exists and check_content:
                # Yippie! a grib2 file exists
                one_exists = True

                # Define the filename we want to save the data as
                outFile = _outFile_from_url(url, save_dir / f'{DATE:%Y%m%d}')

                # Download the data
                #------------------
                if searchString in [None, ':']:
                    if dryrun:
                        if not overwrite and outFile.exists():
                            if verbose: print(f'🌵 Dry Run! Already have file for --> {outFile}')
                            all_files.append(None)
                        else:
                            if verbose: print(f'🌵 Dry Run! Would get from [{source}] {url} --> {outFile}')
                            all_files.append(None)
                    else:
                        # Download the full file.
                        if not overwrite and outFile.exists():
                            if verbose: print(f'🌉 Already have file for --> {outFile}')
                            all_files.append(outFile)
                        else:
                            urllib.request.urlretrieve(url, outFile, _reporthook)
                            all_files.append(outFile)
                            if verbose: print(f'✅ Success! Downloaded from [{source}] {url} --> {outFile}')
                else:
                    # Download a subset of the full file based on the searchString.
                    if verbose: print(f"Download subset from [{source}]:")
                    thisfile = _download_hrrr_subset(url, searchString,
                                                    df_row_urls=row,
                                                    save_dir=outFile.parent,
                                                    dryrun=dryrun,
                                                    verbose=verbose)
                    all_files.append(thisfile)
                all_urls.append(url)
                break # because we downloaded a file and don't need it from another source

        if not one_exists:
            # The URL request is bad or file size is small and the file is not
            # available from any of the sources.
            print()
            print(f'❌ WARNING: Status code {head.status_code}: {head.reason}. Content-Length: {int(head.raw.info()["Content-Length"]):,} bytes')
            print(f"❌ Could not download file for [{model}] [{field}]. Tried to get the following:")
            for i, url in enumerate(row, start=1):
                print(f'    {i}: {url}')
            print('')

        #---------------------------------------------------------
        # Time keeping: *crude* method to estimate remaining time.
        #---------------------------------------------------------
        loop_time += datetime.now() - timer
        mean_dt_per_loop = loop_time/(index+1)
        remaining_loops = n-index-1
        est_rem_time = mean_dt_per_loop * remaining_loops
        if verbose: print(f"🚛💨 Download Progress: [{index+1}/{n} completed] >> Est. Time Remaining {str(est_rem_time):16}\n")
        #---------------------------------------------------------

    print(f"\n🍦 Finished 🍦  Time spent on download: {loop_time}")

    if len(all_files) == 1:
        return all_files[0], all_urls[0]  # return only the one time, not list
    else:
        return np.array(all_files), np.array(all_urls)  # return the list of file names and URLs

#=======================================================================
#=======================================================================

def xhrrr(DATE, searchString, fxx=0, *,
          DATE_is_valid_time=False,
          remove_grib2=True,
          add_crs=True,
          backend_kwargs={},
          **download_kwargs):
    """
    Download HRRR data and return as an xarray Dataset (or Datasets)

    You may only request one `DATE` and one `fxx` (forecast lead time).

    .. note:: 
        See https://github.com/ecmwf/cfgrib/issues/187 for why there is
        a problem with reading multiple accumulated precipitation
        fields when searchString=':APCP:'.

    Parameters
    ----------
    DATE : datetime
        A single datetime object.
    searchString: string
        A string representing a field or fields from the GRIB2 file.
        See more details in ``download_hrrr`` docstring.

        Some examples:

        ============================= ===============================================
        ``searchString``              Messages that will be downloaded
        ============================= ===============================================
        ':TMP:2 m'                    Temperature at 2 m.
        ':TMP:'                       Temperature fields at all levels.
        ':UGRD:.* mb'                 U Wind at all pressure levels.
        ':500 mb:'                    All variables on the 500 mb level.
        ':APCP:'                      All accumulated precipitation fields.
        ':APCP:surface:0-[1-9]*'      Accumulated precip since initialization time
        ':APCP:surface:[1-9]*-[1-9]*' Accumulated precip over last hour
        ':UGRD:10 m'                  U wind component at 10 meters.
        ':(U|V)GRD:'                  U and V wind component at all levels.
        ':.GRD:'                      (Same as above)
        ':(TMP|DPT):'                 Temperature and Dew Point for all levels .
        ':(TMP|DPT|RH):'              TMP, DPT, and Relative Humidity for all levels.
        ':REFC:'                      Composite Reflectivity
        ':surface:'                   All variables at the surface.
        ============================= ===============================================

    fxx : int
        Forecast lead time. Default will get the analysis, F00.
    DATE_is_valid_time: bool
        False - (default) The DATE argument represents the model
                initialization datetime.
        True  - The DATE argument represents the model valid time.
                This is handy when you want a specific forecast leadtime
                that is valid at a certian date.
    remove_grib2 : bool
        True  - (default) Delete the GRIB2 file after reading into a Dataset.
                This requires a copy to memory, so it might slow things down.
        False - Keep the GRIB2 file downloaded.
                This might be a better option performance-wise, because it
                does not need to copy the data but keeps the file on disk.
                You would be responsible for removing files when you don't
                need them.
    add_crs : bool
        True  - (default) Add the Cartopy coordinate reference system (crs)
                projection as an attribute to the Dataset.
    **download_kwargs :
        Any other key word argument accepted by ``download_hrrr``.
            - model : {'hrrr', 'hrrrak', 'hrrrX'}
            - field : {'sfc', 'prs', 'nat', 'subh'}
            - save_dir : pathlib.Path
            - download_source_priority : a list of download sources
            - dryrun : bool
            - verbose : bool
    """
    # Convert DATE input to a pandas datetime (Pandas can parse some strings as dates.)
    DATE = pd.to_datetime(DATE)

    inputs = locals()  

    assert not hasattr(DATE, '__len__'), "`DATE` must be a single datetime, not a list."
    assert not hasattr(fxx, '__len__'), "`fxx` must be a single integer, not a list."

    if DATE_is_valid_time:
        # Change DATE to the model run initialization DATE so that when we take
        # into account the forecast lead time offset, the the returned data
        # be valid for the DATE the user requested.
        DATE = DATE - timedelta(hours=fxx)

    # Download the GRIB2 file
    grib2file, url = download_hrrr(DATE, searchString, fxx=fxx, **download_kwargs)

    # Some extra backend kwargs for cfgrib
    backend_kwargs.setdefault('indexpath', '')
    backend_kwargs.setdefault('read_keys', ['parameterName', 'parameterUnits', 'stepRange'])
    backend_kwargs.setdefault('errors', 'raise')

    # Use cfgrib.open_datasets, just in case there are multiple "hypercubes"
    # for what we requested.
    H = cfgrib.open_datasets(grib2file, backend_kwargs=backend_kwargs)

    # Create a cartopy projection object
    if add_crs:
        crs = get_crs(H[0])

    for ds in H:
        ds.attrs['history'] = inputs
        ds.attrs['url'] = url

        # CF 1.8 map projection information for the HRRR model
        # http://cfconventions.org/Data/cf-conventions/cf-conventions-1.8/cf-conventions.html#_lambert_conformal

        ##
        ## I'm not sure why, but when I assign an attribute for the main Dataset
        ## with a cartopy.crs and then copy it (for the remove_grib2 case),
        ## the cartopy.crs is reset to the default proj4_params.
        ## This isn't an issue when I set cartopy.crs as an attribute for
        ## a variable (DataArray).
        ## For this reason, I will return the 'crs' as an attribute for
        ## each variable's DataArray. Later, I will add the 'crs' as
        ## a Dataset attribute for convenience.
        ##

        ds.attrs['grid_mapping_name'] = 'lambert_conformal_conic'
        ds.attrs['standard_parallel'] = (38.5, 38.5)
        ds.attrs['longitude_of_central_meridian'] = 262.5
        ds.attrs['latitude_of_projection_origin'] = 38.5

        # This is redundant, but I want every variable to also have the
        # map projection information...
        for var in list(ds):
            ds[var].attrs['grid_mapping_name'] = 'lambert_conformal_conic'
            ds[var].attrs['standard_parallel'] = (38.5, 38.5)
            ds[var].attrs['longitude_of_central_meridian'] = 262.5
            ds[var].attrs['latitude_of_projection_origin'] = 38.5
            if add_crs:
                ds[var].attrs['crs'] = crs

    if remove_grib2:
        # Load the data to memory before removing the file
        H = [ds.load() for ds in H]
        # Ok, now we can remove the grib2 file
        os.remove(grib2file)

    if len(H) == 1:
        H = H[0]
        # Add the cartopy map projection object as an attribute, for convenience.
        H.attrs['crs'] = crs
    else:
        warnings.warn('⚠ ALERT! Could not load grib2 data into a single '
                      'xarray Dataset. You might consider refining your '
                      '`searchString` if you are getting data you do not need.')
        # Add the cartopy map projection object as an attribute, for convenience.
        for i in H:
            i.attrs['crs'] = crs

    return H



## Brian Blaylock
## May 3, 2021

"""
============
Herbie Tools
============
"""
from datetime import datetime, timedelta
import pandas as pd
import xarray as xr

from herbie.archive import Herbie


def bulk_download(
    DATES,
    searchString=None,
    *,
    fxx=range(0, 1),
    model="hrrr",
    product="sfc",
    priority=None,
    verbose=True,
):
    """
    Bulk download GRIB2 files from file source to the local machine.

    Iterates over a list of datetimes (DATES) and forecast lead times (fxx).

    Parameters
    ----------
    DATES : list
        List of datetimes
    searchString : None or str
        If None, download the full file. If string, use regex to search
        index files for variables and levels of interest and only
        download the matched GRIB messages.
    fxx : int or list
        List of forecast lead times to download. Default only downloads model analysis.
    model : {'hrrr', 'hrrrak', 'rap'}
        Model to download.
    product : {'sfc', 'prs', 'nat', 'subh'}
        Variable products file to download. Not needed for RAP model.
    """
    if isinstance(DATES, (str, pd.Timestamp)) or hasattr(DATES, "strptime"):
        DATES = [DATES]
    if isinstance(fxx, int):
        fxx = [fxx]

    kw = dict(model=model, product=product)
    if priority is not None:
        kw["priority"] = priority

    # Locate the file sources
    print("👨🏻‍🔬 Check which requested files exists")
    grib_sources = [Herbie(d, fxx=f, **kw) for d in DATES for f in fxx]

    loop_time = timedelta()
    n = len(grib_sources)

    print("\n🌧 Download requested data")
    for i, g in enumerate(grib_sources):
        timer = datetime.now()
        g.download(searchString=searchString)

        # ---------------------------------------------------------
        # Time keeping: *crude* method to estimate remaining time.
        # ---------------------------------------------------------
        loop_time += datetime.now() - timer
        mean_dt_per_loop = loop_time / (i + 1)
        remaining_loops = n - i - 1
        est_rem_time = mean_dt_per_loop * remaining_loops
        if verbose:
            print(
                f"🚛💨 Download Progress: [{i+1}/{n} completed] >> Est. Time Remaining {str(est_rem_time):16}\n"
            )
        # ---------------------------------------------------------

    requested = len(grib_sources)
    completed = sum([i.grib is None for i in grib_sources])
    print(f"🍦 Done! Downloaded [{completed}/{requested}] files. Timer={loop_time}")

    return grib_sources


def xr_concat_sameRun(DATE, searchString, fxx=range(0, 18)):
    """
    Load and concatenate xarray objects by forecast lead time for the same run.

    Parameters
    ----------
    DATE : pandas-parsable datetime
        A datetime that represents the model initialization time.
    searchString : str
        Variable fields to load. This really only works if the search
        string returns data on the same hyper cube.
    fxx : list of int
        List of forecast lead times, in hours, to concat together.
    """
    Hs_to_cat = [Herbie(DATE, fxx=f).xarray(searchString) for f in fxx]
    return xr.concat(Hs_to_cat, dim="f")


def xr_concat_sameLead(DATES, searchString, fxx=0, DATE_is_valid_time=True):
    """
    Load and concatenate xarray objects by model initialization date for the same lead time.

    Parameters
    ----------
    DATES : list of pandas-parsable datetime
        Datetime that represents the model valid time.
    searchString : str
        Variable fields to load. This really only works if the search
        string returns data on the same hyper cube.
    fxx : int
        The forecast lead time, in hours.
    """
    Hs_to_cat = [
        Herbie(DATE, fxx=fxx, DATE_is_valid_time=DATE_is_valid_time).xarray(
            searchString
        )
        for DATE in DATES
    ]
    return xr.concat(Hs_to_cat, dim="t")



## Added by Brian Blaylock
## July 26, 2021 

"""
A Herbie template for the HRRR model.

Because the file path to GRIB2 model data is predictable, we can 
template the download URL for model output. Follow this template for 
writing your own template file for any model with GRIB2 files available
via https.

Requirements
------------
1. Model must be available via https
2. Preferably, an .idx file should be available.
3. URL must be consistent across time and products.

Properties
----------
DESCRIPTION : str
    A description of the model. Give the full name and the 
    domain, if relevant.
DETAILS : dict
    Some additional details about the model. Provide links
    to web documentation.
PRODUCTS : dict
    Models usually have different product types. The keys are
    used in building the GRIB2 source URL.
    ORDER MATTERS -- If product is None, then Herbie uses the first
    as default.
SOURCES : dict
    Build the URL for the GRIB2 file for different sources.
    The parameters are from arguments passed into the 
    ``herbie.archive.Herbie()`` class.
    ORDER MATTERS -- If priority is None, then Herbie searches the
    sources in the order given here.
LOCALFILE : str
    The local file to save the model output. The file will be saved in
    ``save_dir/model/YYYYmmdd/localFile.grib2``
    It is sometimes necessary to add details to maintain unique 
    filenames (e.g., rrfs needs to have the member number in LOCALFILE).
"""

class hrrr:
    def template(self):
        self.DESCRIPTION = 'High-Resolution Rapid Refresh - CONUS'
        self.DETAILS = {
            'NOMADS product description': 'https://www.nco.ncep.noaa.gov/pmb/products/hrrr/',
            'University of Utah HRRR archive': 'http://hrrr.chpc.utah.edu/',
            }
        self.PRODUCTS = {
            'sfc': "2D surface level fields; 3-km resolution",
            'prs': "3D pressure level fields; 3-km resolution",
            'nat': "Native level fields; 3-km resolution",
            'subh': "Subhourly grids; 3-km resolution"
        }
        self.SOURCES = {
            'aws': f'https://noaa-hrrr-bdp-pds.s3.amazonaws.com/hrrr.{self.date:%Y%m%d}/conus/hrrr.t{self.date:%H}z.wrf{self.product}f{self.fxx:02d}.grib2',
            'nomads': f'https://nomads.ncep.noaa.gov/pub/data/nccf/com/hrrr/prod/hrrr.{self.date:%Y%m%d}/conus/hrrr.t{self.date:%H}z.wrf{self.product}f{self.fxx:02d}.grib2',
            'google': f'https://storage.googleapis.com/high-resolution-rapid-refresh/hrrr.{self.date:%Y%m%d}/conus/hrrr.t{self.date:%H}z.wrf{self.product}f{self.fxx:02d}.grib2',
            'azure': f'https://noaahrrr.blob.core.windows.net/hrrr/hrrr.{self.date:%Y%m%d}/conus/hrrr.t{self.date:%H}z.wrf{self.product}f{self.fxx:02d}.grib2',
            'pando': f'https://pando-rgw01.chpc.utah.edu/{self.model}/{self.product}/{self.date:%Y%m%d}/{self.model}.t{self.date:%H}z.wrf{self.product}f{self.fxx:02d}.grib2',
            'pando2': f'https://pando-rgw02.chpc.utah.edu/{self.model}/{self.product}/{self.date:%Y%m%d}/{self.model}.t{self.date:%H}z.wrf{self.product}f{self.fxx:02d}.grib2',
        }
        self.LOCALFILE = f"{self.get_remoteFileName}"


## Added by Brian Blaylock
## July 26, 2021

class rap:
    def template(self):
        self.DESCRIPTION = 'Rapid Refresh'
        self.DETAILS = {
            'nomads product description': 'https://www.nco.ncep.noaa.gov/pmb/products/rap',
        }
        self.PRODUCTS = {
            'wrfprs.': 'Full domain Pressure Levels; 13-km',
            'wrfnat.': 'Full domain Native Levels; 13-km',
            'awp130pgrb': 'CONUS Pressure levels; 13-km resolution',
            'awp252pgrb': 'CONUS Pressure levels; 20-km resolution',
            'awp236pgrb': 'CONUS Pressure levels; 40-km resolution',
            'awp130bgrb': 'CONUS Native levels; 13-km resolution',
            'awp252bgrb': 'CONUS Native levels; 20-km resolution',
            'awip32': 'NOAMHI - High-Resolution North American Master Grid; 32-km resolution',
            'awp242': 'Alaska Quadruple Resolution Pressure levels; 11-km resolution',
            'awp200': 'Puerto Rico Pressure levels; 16-km resolution',
            'awp243': 'Eastern North America Pressure levels, 0.4 degree resolution',
            'wrfmsl': 'WRFMSL; 13-km resolution',
        }
        self.SOURCES = {
            'nomads': f'https://nomads.ncep.noaa.gov/pub/data/nccf/com/rap/prod/rap.{self.date:%Y%m%d}/rap.t{self.date:%H}z.{self.product}f{self.fxx:02d}.grib2',
            'aws': f'https://noaa-rap-pds.s3.amazonaws.com/rap.{self.date:%Y%m%d}/rap.t{self.date:%H}z.{self.product}f{self.fxx:02d}.grib2',
            'google': f'https://storage.googleapis.com/rapid-refresh/rap.{self.date:%Y%m%d}/rap.t{self.date:%H}z.{self.product}f{self.fxx:02d}.grib2',
            'azure': f'https://noaarap.blob.core.windows.net/rap/rap.{self.date:%Y%m%d}/rap.t{self.date:%H}z.{self.product}f{self.fxx:02d}.grib2',
        }
        self.LOCALFILE = f"{self.get_remoteFileName}"


## Brian Blaylock
## May 13, 2021

import matplotlib.pyplot as plt
import matplotlib.patheffects as path_effects


def HerbieColors():
    return dict(
        body="#f0ead2", red="#88211b", blue="#0c3576", white="#ffffff", black="#000000"
    )


def HerbieLogo(white_line=False):
    """Logo of Herbie The Love Bug"""
    colors = HerbieColors()

    plt.figure(figsize=[5, 5], facecolor=colors["body"])

    plt.axis([-10, 10, -10, 10])

    if white_line:
        plt.axvline(4, lw=40, color=colors["white"])
    plt.axvline(2.5, lw=20, color=colors["red"])
    plt.axvline(5.5, lw=40, color=colors["blue"])

    c = plt.Circle((0, 0), radius=6, ec="k", fc="w", zorder=10, linewidth=3)
    plt.gca().add_artist(c)

    plt.text(
        0,
        0,
        "53",
        fontsize=110,
        fontweight="bold",
        va="center_baseline",
        ha="center",
        zorder=11,
    )

    plt.gca().set_facecolor(colors["body"])
    plt.gca().margins(0)

    plt.gca().get_xaxis().set_visible(False)
    plt.gca().get_yaxis().set_visible(False)

    plt.gca().spines["bottom"].set_visible(False)
    plt.gca().spines["top"].set_visible(False)
    plt.gca().spines["left"].set_visible(False)
    plt.gca().spines["right"].set_visible(False)

    return plt.gca()


def HerbieLogo2(white_line=False, text_color="body", text_stroke="black"):
    """Logo of Herbie The Love Bug"""
    colors = HerbieColors()

    plt.figure(figsize=[5, 3], facecolor=colors["body"])

    plt.axis([1.5, 20, -10, 10])

    if white_line:
        plt.axvline(4, lw=40, color=colors["white"])
    plt.axvline(2.5, lw=20, color=colors["red"])
    plt.axvline(5.5, lw=40, color=colors["blue"])

    if text_color in colors:
        text_color = colors[text_color]
    if text_stroke in colors:
        text_stroke = colors[text_stroke]

    text = plt.text(
        8,
        0,
        "Herbie",
        fontsize=110,
        fontweight="bold",
        color=text_color,
        va="center_baseline",
        ha="left",
        zorder=11,
    )

    if text_stroke is not None:
        text.set_path_effects(
            [
                path_effects.Stroke(linewidth=3, foreground=text_stroke),
                path_effects.Normal(),
            ]
        )

    plt.gca().set_facecolor(colors["body"])
    plt.gca().margins(0)

    plt.gca().get_xaxis().set_visible(False)
    plt.gca().get_yaxis().set_visible(False)

    plt.gca().spines["bottom"].set_visible(False)
    plt.gca().spines["top"].set_visible(False)
    plt.gca().spines["left"].set_visible(False)
    plt.gca().spines["right"].set_visible(False)

    return plt.gca()



## Brian Blaylock
## April 23, 2021

"""
==================================
HRRR-B Extension: xarray accessors
==================================

Extend the xarray capabilities with a custom accessor.
http://xarray.pydata.org/en/stable/internals.html#extending-xarray

"""
import warnings

import numpy as np
from paint.radar import cm_reflectivity
import pandas as pd
import xarray as xr
import matplotlib.pyplot as plt

from toolbox.cartopy_tools import common_features, pc
from paint.standard2 import cm_tmp, cm_dpt, cm_rh, cm_wind, cm_pcp
from paint.radar2 import cm_reflectivity


@xr.register_dataset_accessor("hrrrb")
class hrrrb_accessor:
    def __init__(self, xarray_obj):
        self._obj = xarray_obj
        self._center = None

    @property
    def center(self):
        """Return the geographic center point of this dataset."""
        if self._center is None:
            # we can use a cache on our accessor objects, because accessors
            # themselves are cached on instances that access them.
            lon = self._obj.latitude
            lat = self._obj.longitude
            self._center = (float(lon.mean()), float(lat.mean()))
        return self._center

    def plot(self, ax=None, common_features_kw={}, **kwargs):
        """Plot data on a map."""
        ds = self._obj

        _level_units = dict(
            adiabaticCondensation='adiabatic condensation',
            atmosphere='atmosphere',
            atmosphereSingleLayer='atmosphere single layer',
            boundaryLayerCloudLayer='boundary layer cloud layer',
            cloudBase='cloud base',
            cloudCeiling='cloud ceiling',
            cloudTop='cloud top',
            depthBelowLand='m',
            equilibrium='equilibrium',
            heightAboveGround='m',
            heightAboveGroundLayer='m',
            highCloudLayer='high cloud layer',
            highestTroposphericFreezing='highest tropospheric freezing',
            isobaricInhPa='hPa',
            isobaricLayer='hPa',
            isothermZero='0 C',
            isothermal='K',
            level='m',
            lowCloudLayer='low cloud layer',
            meanSea='MSLP',
            middleCloudLayer='middle cloud layer',
            nominalTop='nominal top',
            pressureFromGroundLayer='Pa',
            sigma='sigma',
            sigmaLayer='sigmaLayer',
            surface='surface',
        )

        for var in ds.data_vars:
            print('cfgrib variable:', var)
            print('GRIB_cfName', ds[var].GRIB_cfName)
            print('GRIB_cfVarName', ds[var].GRIB_cfVarName)
            print('GRIB_name', ds[var].GRIB_name)
            print('GRIB_units', ds[var].GRIB_units)
            print('GRIB_typeOfLevel', ds[var].GRIB_typeOfLevel)
            print()
            ds[var].attrs['units'] = ds[var].attrs['units'].replace('**-1', '$^{-1}$').replace('**-2', '$^{-2}$')
            
            dpi = common_features_kw.pop('dpi', 150)
            figsize = common_features_kw.pop('figsize', [10,5])
            fig, ax = plt.subplots(1,1, subplot_kw=dict(projection=ds.crs), dpi=dpi, figsize=figsize)
            
            default = dict(scale='50m', ax=ax, crs=ds.crs, STATES=True, BORDERS=True)
            common_features_kw = {**default, **common_features_kw}
            common_features(**common_features_kw)

            title = ''
            kwargs = {}
            cbar_kwargs = dict(pad=0.01)

            if ds[var].GRIB_cfVarName in ['d2m', 'dpt']:
                ds[var].attrs['GRIB_cfName'] = 'dew_point_temperature'

            if ds[var].GRIB_cfName == 'air_temperature':
                kwargs = {**cm_tmp().cmap_kwargs, **kwargs}
                cbar_kwargs = {**cm_tmp().cbar_kwargs, **cbar_kwargs}
                if ds[var].GRIB_units == 'K':
                    ds[var] -= 273.15
                    ds[var].attrs['GRIB_units'] = 'C'
                    ds[var].attrs['units'] = 'C'

            elif ds[var].GRIB_cfName == 'dew_point_temperature':
                kwargs = {**cm_dpt().cmap_kwargs, **kwargs}
                cbar_kwargs = {**cm_dpt().cbar_kwargs, **cbar_kwargs}
                if ds[var].GRIB_units == 'K':
                    ds[var] -= 273.15
                    ds[var].attrs['GRIB_units'] = 'C'
                    ds[var].attrs['units'] = 'C'
                
            elif ds[var].GRIB_name == 'Total Precipitation':
                title = '-'.join([f'F{int(i):02d}' for i in ds[var].GRIB_stepRange.split('-')])
                ds[var] = ds[var].where(ds[var]!=0)
                kwargs = {**cm_pcp().cmap_kwargs, **kwargs}
                cbar_kwargs = {**cm_pcp().cbar_kwargs, **cbar_kwargs}

            elif ds[var].GRIB_name == 'Maximum/Composite radar reflectivity':
                ds[var] = ds[var].where(ds[var]>=0)
                cbar_kwargs = {**cm_reflectivity().cbar_kwargs, **cbar_kwargs}
                kwargs = {**cm_reflectivity().cmap_kwargs, **kwargs}

            elif ds[var].GRIB_cfName == 'relative_humidity':
                cbar_kwargs = {**cm_rh().cbar_kwargs, **cbar_kwargs}
                kwargs = {**cm_rh().cmap_kwargs, **kwargs}
                
            else:
                cbar_kwargs = {**dict(label=f"{ds[var].GRIB_parameterName.strip().title()} ({ds[var].units})"), **cbar_kwargs}            
            
            p = ax.pcolormesh(ds.longitude, ds.latitude, ds[var], transform=pc, **kwargs)
            plt.colorbar(p, ax=ax, **cbar_kwargs)

            VALID = pd.to_datetime(ds.valid_time.data).strftime('%H:%M UTC %d %b %Y')
            RUN = pd.to_datetime(ds.time.data).strftime('%H:%M UTC %d %b %Y')
            FXX = f"F{pd.to_timedelta(ds.step.data).total_seconds()/3600:02.0f}"
            
            level_type = ds[var].GRIB_typeOfLevel
            if level_type in _level_units:
                level_units = _level_units[level_type]
            else:
                level_units = 'unknown'

            if level_units.lower() in ['surface', 'atmosphere']:
                level = f'{title} {level_units}'
            else:
                level = f'{ds[var][level_type].data:g} {level_units}'

            ax.set_title(f"Run: {RUN} {FXX}", loc='left', fontfamily='monospace', fontsize='x-small')
            ax.set_title(f'HRRR {level}', loc='center', fontweight='semibold')
            ax.set_title(f"Valid: {VALID}", loc='right', fontfamily='monospace', fontsize='x-small')

            # Set extent (could do this more efficiently by storing the data elsewhere)
            new = ds.crs.transform_points(pc, ds.longitude.data, ds.latitude.data)
            LONS = new[:,:,0]
            LATS = new[:,:,1]
            ax.set_extent([LONS.min(), LONS.max(), LATS.min(), LATS.max()], crs=ds.crs)
        
        return ax



## Brian Blaylock
## September 28, 2020

"""
==========
HRRR Plots
==========
"""
import warnings
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
import cartopy.feature as feature
import cartopy.io.img_tiles as cimgt
from shapely.geometry import Polygon
import xarray as xr
import numpy as np
try:
    from metpy.plots import USCOUNTIES
except:
    print('metpy package not found/imported, so you can not plot US Counties')
    

def simple_plot(ds, ax=None):
    """
    Make a simple plot for each variable.
    """
    if isinstance(ds, list):
        # A little recursion, for the case when get_hrrr returns a list
        # of xr.Datasets...
        for i in ds:
            simple_plot(i, ax=ax)
        return None

    for var in ds.data_vars:
        fig = plt.figure()
        ax = plt.axes(projection=ds.crs)
        ds[var].plot(x='longitude', y='latitude', 
                     ax=ax, transform=ccrs.PlateCarree())
        ax.coastlines()
    return None


## Brian Blaylock
## September 28, 2020

"""
==========================
Tools for the HRRR Archive
==========================

to_180
    For longitude values to be from -180 W to 180 E (not 0-360 E).
get_crs
    Get cartopy projection object from xarray.Dataset
pluck_points
    Pluck values at specific latitude/longitude points.
"""

import numpy as np
import xarray as xr
import cartopy.crs as ccrs

import hrrrb.archive as ha
import multiprocessing

def to_180(lon):
    """
    Wrap longitude from degrees [0, 360] to degrees [-180, 180].

    An alternative method is
        lon[lon>180] -= 360

    Parameters
    ----------
    lon : array_like
        Longitude values
    """
    lon = np.array(lon)
    lon = (lon + 180) % 360 - 180
    return lon

def get_crs(ds):
    """
    Get the cartopy coordinate reference system (crs) from a cfgrib xarray Dataset
    
    Projection in formation is from the grib2 message for each variable.
    
    Parameters
    ----------
    ds : xarray.Dataset
        An xarray.Dataset from a GRIB2 file opened by the cfgrib engine.
    """
    # Get projection from the attributes of HRRR xr.Dataset
    # CF 1.8 map projection information for the HRRR model
    # http://cfconventions.org/Data/cf-conventions/cf-conventions-1.8/cf-conventions.html#_lambert_conformal

    
    if isinstance(ds, list):
        # The case when get_hrrr returns a list of xr.Datasets
        return get_crs(ds[0])


    if isinstance(ds, xr.Dataset):
        attrs = ds[list(ds)[0]].attrs
    if isinstance(ds, xr.DataArray):
        attrs = ds.attrs
    if attrs['GRIB_gridType'] == 'lambert':
        lc_hrrr_kwargs = {
            'globe': ccrs.Globe(ellipse='sphere'),
            'central_latitude': attrs['GRIB_LaDInDegrees'],
            'central_longitude': attrs['GRIB_LoVInDegrees'],
            'standard_parallels': (attrs['GRIB_Latin1InDegrees'],\
                                   attrs['GRIB_Latin2InDegrees'])}
        lc = ccrs.LambertConformal(**lc_hrrr_kwargs)
        return lc
    else:
        warnings.warn('GRIB_gridType is not "lambert".')
        return None

def pluck_points(ds, points, names=None, dist_thresh=10_000, verbose=True):
    """
    Pluck values at point nearest a give list of latitudes and longitudes pairs.
    
    Uses a nearest neighbor approach to get the values. The general 
    methodology is illustrated in this 
    `GitHub Notebook <https://github.com/blaylockbk/pyBKB_v3/blob/master/demo/Nearest_lat-lon_Grid.ipynb>`_.
        
    Parameters
    ----------
    ds : xarray.Dataset
        The Dataset should include coordinates for both 'latitude' and 
        'longitude'.
    points : tuple or list of tuples
        The latitude and longitude (lat, lon) coordinate pair (as a tuple)
        for the points you want to pluck from the gridded Dataset. 
        A list of tuples may be given to return the values from multiple points.
    names : list
        A list of names for each point location (i.e., station name).
        None will not append any names. names should be the same
        length as points.
    dist_thresh: int or float
        The maximum distance (m) between a plucked point and a matched point.
        Default is 10,000 m. If the distance is larger than this, the point
        is disregarded.
    
    Returns
    -------
    The Dataset values at the points nearest the requested lat/lon points.
    """
    if 'lat' in ds:
        ds = ds.rename(dict(lat='latitude', lon='longitude'))
    
    if isinstance(points, tuple):
        # If a tuple is give, turn into a one-item list.
        points = [points]

    if names is not None:
        assert len(points) == len(names), '`points` and `names` must be same length.'
        
    # Find the index for the nearest points
    xs = []  # x index values
    ys = []  # y index values 
    for point in points:
        assert len(point) == 2, "``points`` should be a tuple or list of tuples (lat, lon)"
        
        p_lat, p_lon = point

        # Force longitude values to range from -180 to 180 degrees.
        p_lon = to_180(p_lon)
        ds['longitude'][:] = to_180(ds.longitude)

        # Find absolute difference between requested point and the grid coordinates.
        abslat = np.abs(ds.latitude - p_lat)
        abslon = np.abs(ds.longitude - p_lon)

        # Create grid of the maximum values of the two absolute grids
        c = np.maximum(abslon, abslat)

        # Find location where lat/lon minimum absolute value intersects
        if ds.latitude.dims == ('y', 'x'):
            y, x = np.where(c == np.min(c))
        elif ds.latitude.dims == ('x', 'y'):
            x, y = np.where(c == np.min(c))
        else:
            raise ValueError(f"Sorry, I do not understand dimensions {ds.latitude.dims}. Expected ('y', 'x')" )
        xs.append(x[0])
        ys.append(y[0])

    #===================================================================
    # Select Method 1:
    # This method works, but returns more data than you ask for.
    # It returns an NxN matrix where N is the number of points, 
    # and matches each point with each point (not just the coordinate
    # pairs). The points you want will be along the diagonal.
    # I leave this here so I remember not to do this.
    #
    #ds = ds.isel(x=xs, y=ys)   
    #
    #===================================================================
    
    #===================================================================
    # Select Method 2:
    # This is only *slightly* slower, but returns just the data at the
    # points you requested. Creates a new dimension, called 'point'
    ds = xr.concat([ds.isel(x=i, y=j) for i, j in zip(xs, ys)], dim='point')
    #===================================================================
    
    #-------------------------------------------------------------------
    # 📐Approximate the Great Circle distance between matched point and 
    # requested point.
    # Based on https://andrew.hedges.name/experiments/haversine/
    #-------------------------------------------------------------------
    lat1 = np.deg2rad([i[0] for i in points])
    lon1 = np.deg2rad([i[1] for i in points])
    
    lat2 = np.deg2rad(ds.latitude.data)
    lon2 = np.deg2rad(ds.longitude.data)
    
    R = 6373.0 # approximate radius of earth in km
    
    dlon = lon2 - lon1
    dlat = lat2 - lat1

    a = np.sin(dlat / 2)**2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2)**2
    c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))

    distance = R * c * 1000  # converted to meters
    
    # Add the distance values as a coordinate
    ds.coords['distance'] = ('point', distance)
    ds['distance'].attrs = dict(long_name='Distance between requested point and matched grid point', units='m')
    
    #-------------------------------------------------------------------
    #-------------------------------------------------------------------
    
    # Add list of names as a coordinate
    if hasattr(names, '__len__'):
        # Assign the point dimension as the names.
        assert len(ds.point) == len(names), f'`names` must be same length as `points` pairs.'
        ds['point'] = names
    
    ## Print some info about each point:
    if verbose:
        zipped = zip([i[0] for i in points], [i[1] for i in points],
                     ds.latitude.data, ds.longitude.data, ds.distance.data, ds.point.data)
        for plat, plon, glat, glon, d, name in zipped:
            print(f"🔎 Matched requested point [{name}] ({plat:.3f}, {plon:.3f}) to grid point ({glat:.3f}, {glon:.3f})...distance of {d/1000:,.2f} km.")
            if d > dist_thresh:
                print(f'   💀 Point [{name}] Failed distance threshold')

    ds.attrs['x_index'] = xs
    ds.attrs['y_index'] = ys
    
    # Drop points that do not meet the dist_thresh criteria
    failed = ds.distance < dist_thresh
    print(len(failed), failed.data)
    warnings.warn(f' 💀 Dropped {np.sum(failed).data} point(s) that exceeded dist_thresh.')
    ds = ds.where(failed, drop=True)  
    
    return ds

#=======================================================================
#=======================================================================

## WORK IN PROGRESS 🏗 

# A method to download multiple forecasts and read into xarray
def _concat_hrrr(inputs):
    """Multiprocessing Helper"""
    DATE, searchString, fxx, kwargs = inputs
    return ha.xhrrr(DATE, searchString, fxx=fxx, **kwargs)

def concat_hrrr(DATES, searchString, fxxs=range(19), **get_hrrr_kwargs):
    """
    Download variables 
    """
    inputs = [(DATE, searchString, f, get_hrrr_kwargs) for f in fxxs]
    
    cores = np.minimum(len(inputs), multiprocessing.cpu_count())
    print('cores=', cores)
                       
    with multiprocessing.Pool(cores) as p:
        result = p.map(_concat_hrrr, inputs)
        p.close()
        p.join()
    
    return xr.concat(result, dim='lead')




## Brian Blaylock
## April 28, 2021

import warnings
import toml
import os
from pathlib import Path

########################################################################
# Append Path object with my custom expand method so user can use
# environment variables in the config file (e.g., ${HOME}).
def _expand(self):
    """
    Fully expand and resolve the Path with the given environment variables.

    Example
    -------
    >>> Path('$HOME').expand()
    >>> PosixPath('/p/home/blaylock')
    """
    return Path(os.path.expandvars(self)).expanduser().resolve()


Path.expand = _expand

########################################################################
# Herbie configuration file
# Configuration file is save in `~/config/herbie/config.toml`
_config_path = Path("~/.config/herbie/config.toml").expand()

# NOTE: The `\\` is an escape character in TOML. 
# For Windows paths "C:\\user\\"" needs to be "C:\\\\user\\\\""
_save_dir = str(Path('~/data').expand())
_save_dir = _save_dir.replace('\\', '\\\\')


########################################################################
# Default TOML Configuration Values
default_toml = f"""
['default']
model = "hrrr"
fxx = 0
priority = ['aws', 'nomads', 'google', 'azure', 'pando', 'pando2', 'aws-old']
save_dir = "{_save_dir}"
overwrite = false
verbose = true
"""

########################################################################
# If a config file isn't found, make one
if not _config_path.exists():
    with open(_config_path, "w") as f:
        toml_string = toml.dump(toml.loads(default_toml), f)
    print(f"⚙ Created config file [{_config_path}] with default values.")

########################################################################
# Read the config file
config = toml.load(_config_path)

config["default"]["save_dir"] = Path(config["default"]["save_dir"]).expand()



## Added by Brian Blaylock
## July 26, 2021

class hrrrak:
    def template(self):
        self.DESCRIPTION = 'High-Resolution Rapid Refresh - Alaska'
        self.DETAILS = {
            'nomads product description': 'https://www.nco.ncep.noaa.gov/pmb/products/hrrr',
        }
        self.PRODUCTS = {
            'prs': "3D pressure level fields; 3-km resolution",
            'sfc': "2D surface level fields; 3-km resolution",
            'nat': "Native level fields; 3-km resolution",
            'subh': "Subhourly grids; 3-km resolution"
        }
        self.SOURCES = {
            'nomads': f'https://nomads.ncep.noaa.gov/pub/data/nccf/com/hrrr/prod/hrrr.{self.date:%Y%m%d}/alaska/hrrr.t{self.date:%H}z.wrf{self.product}f{self.fxx:02d}.ak.grib2',
            'aws': f'https://noaa-hrrr-bdp-pds.s3.amazonaws.com/hrrr.{self.date:%Y%m%d}/alaska/hrrr.t{self.date:%H}z.wrf{self.product}f{self.fxx:02d}.ak.grib2',
            'google': f'https://storage.googleapis.com/high-resolution-rapid-refresh/hrrr.{self.date:%Y%m%d}/alaska/hrrr.t{self.date:%H}z.wrf{self.product}f{self.fxx:02d}.ak.grib2',
            'azure': f'https://noaahrrr.blob.core.windows.net/hrrr/hrrr.{self.date:%Y%m%d}/alaska/hrrr.t{self.date:%H}z.wrf{self.product}f{self.fxx:02d}.ak.grib2',
            'pando': f'https://pando-rgw01.chpc.utah.edu/{self.model}/{self.product}/{self.date:%Y%m%d}/{self.model}.t{self.date:%H}z.wrf{self.product}f{self.fxx:02d}.grib2',
            'pando2': f'https://pando-rgw02.chpc.utah.edu/{self.model}/{self.product}/{self.date:%Y%m%d}/{self.model}.t{self.date:%H}z.wrf{self.product}f{self.fxx:02d}.grib2',
        }
        self.LOCALFILE = f"{self.get_remoteFileName}"


#!/usr/bin/env python3

## Brian Blaylock
## May 3, 2021

"""
===============================
Herbie: Retrieve NWP Model Data
===============================

Herbie is your model output download assistant with a mind of his own!
Herbie might look small on the outside, but he has a big heart on the 
inside and will get you to the
`finish line <https://www.youtube.com/watch?v=4XWufUZ1mxQ&t=189s>`_.
Happy racing! 🏎🏁

`📔 Documentation <https://blaylockbk.github.io/Herbie/_build/html/>`_

With Herbie's API, you can search and download GRIB2 model output files
from different archive sources for the High-Resolution Rapid Refresh 
(HRRR) HRRR-Alaska, Rapid Refresh (RAP), Global Forecast System (GFS), 
and others.

Herbie looks for GRIB2 model output data from NOMADS, NOAA's Big Data 
Project partners (Amazon Web Services, Google Cloud Platform, and 
Microsoft Azure), and the CHPC Pando archive at the University of Utah.

Herbie supports subsetting of GRIB2 files by individual GRIB
messages (i.e. variable and level) when the index (.idx) file exist and
help you open them with xarray/cfgrib.

Herbie is extendable to support other models. Simply create a template
file in the ``herbie/models`` directory and make a pull-request.

For more details, see https://blaylockbk.github.io/Herbie/_build/html/user_guide/data_sources.html

.. note:: Updates since the ``Herbie 0.0.5`` release

    - TODO: Rename 'searchString' to 'subset' (and rename subset function)
    - TODO: Create .idx file if wgrib2 is installed (linux only) when index file doesn't exist
    - TODO: add `idx_to_df()` and `df_to_idx()` methods.
    - TODO: clean up document examples. It's kind of scattered now.

"""
import os
import warnings
from datetime import datetime, timedelta
import urllib.request
import requests
import cfgrib
import pandas as pd

# Path imported from __init__ because it has my custom `expand()` method
from . import Path

# NOTE: These config dict values are retrieved from __init__ and read
# from the file ${HOME}/.config/herbie/config.toml
from . import config

import herbie.models as models_template

try:
    # Load custom xarray accessors
    import herbie.accessors
except:
    warnings.warn(
        "herbie xarray accessors could not be imported."
        "You are probably missing the Carpenter_Workshop."
        "If you want to use these functions, try"
        "`pip install git+https://github.com/blaylockbk/Carpenter_Workshop.git`"
    )
    pass


def _searchString_help():
    """Help/Error Message for `searchString`"""
    msg = [
        "\nUse regular expression to search for lines in the .idx file",
        "Here are some examples you can use for `searchString`",
        "  ============================= ===============================================",
        "  ``searchString``              Messages that will be downloaded",
        "  ============================= ===============================================",
        "  ':TMP:2 m'                    Temperature at 2 m.",
        "  ':TMP:'                       Temperature fields at all levels.",
        "  ':UGRD:.* mb'                 U Wind at all pressure levels.",
        "  ':500 mb:'                    All variables on the 500 mb level.",
        "  ':APCP:'                      All accumulated precipitation fields.",
        "  ':APCP:surface:0-[1-9]*'      Accumulated precip since initialization time",
        "  ':APCP:surface:[1-9]*-[1-9]*' Accumulated precip over last hour",
        "  ':UGRD:10 m'                  U wind component at 10 meters.",
        "  ':(U|V)GRD:(10|80) m'         U and V wind component at 10 and 80 m.",
        "  ':(U|V)GRD:'                  U and V wind component at all levels.",
        "  ':.GRD:'                      (Same as above)",
        "  ':(TMP|DPT):'                 Temperature and Dew Point for all levels .",
        "  ':(TMP|DPT|RH):'              TMP, DPT, and Relative Humidity for all levels.",
        "  ':REFC:'                      Composite Reflectivity",
        "  ':surface:'                   All variables at the surface.",
        "  ============================= ===============================================",
        "\nIf you need help with regular expression, search the web",
        "  or look at this cheatsheet: https://www.petefreitag.com/cheatsheets/regex/.",
    ]
    return "\n".join(msg)


class Herbie:
    """
    Locate GRIB2 file at one of the archive sources.
    
    Parameters
    ----------
    date : pandas-parsable datetime
        *Model initialization datetime*.
        If None, then must set ``valid_date``.
    valid_date : pandas-parsable datetime
        Model valid datetime. Must set when ``date`` is None.
    fxx : int
        Forecast lead time in hours. Available lead times depend on
        the model type and model version. Range is model and run
        dependant.
    model : {'hrrr', 'hrrrak', 'rap', 'gfs', 'gfs_wave', 'rrfs', etc.}
        Model name as defined in the models template folder. CASE INSENSITIVE
        Some examples:
        - ``'hrrr'`` HRRR contiguous United States model
        - ``'hrrrak'`` HRRR Alaska model (alias ``'alaska'``)
        - ``'rap'`` RAP model
    product : {'sfc', 'prs', 'nat', 'subh'}
        Output variable product file type. If not specified, will
        use first product in model template file. CASE SENSITIVE.
        For example, the HRRR model has these products:
        - ``'sfc'`` surface fields
        - ``'prs'`` pressure fields
        - ``'nat'`` native fields
        - ``'subh'`` subhourly fields
    member : None or int
        Some ensemble models (e.g. the future RRFS) will need to
        specify an ensemble member.
    priority : list or str
        List of model sources to get the data in the order of
        download priority. CASE INSENSITIVE. Some example data
        sources and the default priority order are listed below.
        - ``'aws'`` Amazon Web Services (Big Data Program)
        - ``'nomads'`` NOAA's NOMADS server
        - ``'google'`` Google Cloud Platform (Big Data Program)
        - ``'azure'`` Microsoft Azure (Big Data Program)
        - ``'pando'`` University of Utah Pando Archive (gateway 1)
        - ``'pando2'`` University of Utah Pando Archive (gateway 2)
    save_dir : str or pathlib.Path
        Location to save GRIB2 files locally. Default save directory
        is set in ``~/.config/herbie/config.cfg``.
    Overwrite : bool
        If True, look for GRIB2 files even if local copy exists.
        If False (default), use the local copy (still need to find
        the idx file).
    """

    def __init__(
        self,
        date=None,
        *,
        valid_date=None,
        model=config["default"].get("model"),
        fxx=config["default"].get("fxx"),
        product=config["default"].get("product"),
        member=config["default"].get("member", 1),
        priority=config["default"].get("priority"),
        save_dir=config["default"].get("save_dir"),
        overwrite=config["default"].get("overwrite", False),
        verbose=config["default"].get("verbose", True),
    ):
        """
        Specify model output and find GRIB2 file at one of the sources.
        """
        self.fxx = fxx

        if date is not None:
            # User supplied `date`, which is the model initialization datetime.
            self.date = pd.to_datetime(date)
            self.valid_date = self.date + timedelta(hours=self.fxx)
        else:
            assert valid_date is not None, "`date` or `valid_date` is required."
            # User supplied `valid_date`, which is the model valid datetime.
            self.valid_date = pd.to_datetime(valid_date)
            self.date = self.valid_date - timedelta(hours=self.fxx)

        self.model = model.lower()
        self.member = member
        self.product = product

        self.priority = priority
        self.save_dir = Path(save_dir).expand()
        self.overwrite = overwrite

        # Get details from the template of the specified model.
        # This attaches the details from the `models.<model>.template`
        # class to this Herbie object.
        # This line is equivalent to `models_template.gfs.template(self)`.
        # We do it this way because the model name is a variable.
        # (see https://stackoverflow.com/a/7936588/2383070 for what I'm doing here)
        getattr(models_template, self.model).template(self)

        if product is None:
            # The user didn't specify a product, so lets use the first
            # product in the model template.
            self.product = list(self.PRODUCTS)[0]
            warnings.warn(f'`product` not specified. Will use ["{self.product}"].')
            # We need to rerun this so the sources have the new product value.
            getattr(models_template, self.model).template(self)

        self.product_description = self.PRODUCTS[self.product]

        # Check the user input
        self._validate()

        # Ok, now we are ready to look for the GRIB2 file at each of the remote sources.
        # self.grib is the first existing GRIB2 file discovered.
        # self.idx is the first existing index file discovered.
        self.grib = None
        self.grib_source = None
        self.idx = None
        self.idx_source = None

        # But first, check if the GRIB2 file exists locally.
        local_copy = self.get_localFilePath()
        if local_copy.exists() and not overwrite:
            self.grib = local_copy
            self.grib_source = "local"
            # NOTE: We will still get the idx files from a remote
            #       because they aren't stored locally.

        # If priority list is set, we want to search SOURCES in that
        # priority order. If priority is None, then search all SOURCES
        # in the order given by the model template file.
        # NOTE: A source from the template will not be used if it is not
        # included in the priority list.
        if self.priority is not None:
            self.SOURCES = {
                key: self.SOURCES[key] for key in self.priority if key in self.SOURCES
            }

        # Ok, NOW we are ready to search for the remote GRIB2 files...
        for source in self.SOURCES:
            if "pando" in source:
                # Sometimes pando returns a bad handshake. Pinging
                # pando first can help prevent that.
                self._ping_pando()

            # Get the file URL for the source and determine if the
            # GRIB2 file and the index file exist. If found, store the
            # URL for the GRIB2 file and the .idx file.
            url = self.SOURCES[source]

            found_grib = False
            found_idx = False
            if self.grib is None and self._check_grib(url):
                found_grib = True
                self.grib = url
                self.grib_source = source
            if self.idx is None and self._check_idx(url):
                found_idx = True
                self.idx = url + ".idx"
                self.idx_source = source

            if verbose:
                msg = (
                    f"Looked in [{source:^10s}] for {self.model.upper()} "
                    f"{self.date:%H:%M UTC %d %b %Y} F{self.fxx:02d} "
                    f"--> ({found_grib=}) ({found_idx=}) {' ':5s}"
                )
                if verbose:
                    print(msg, end="\r", flush=True)

            if all([self.grib is not None, self.idx is not None]):
                # Exit loop early if we found both GRIB2 and idx file.
                break

        # After searching each source, print some info about what we found...
        if verbose:
            if any([self.grib is not None, self.idx is not None]):
                print(
                    f"🏋🏻‍♂️ Found",
                    f"\033[32m{self.date:%Y-%b-%d %H:%M UTC} F{self.fxx:02d}\033[m",
                    f"[{self.model.upper()}] [product={self.product}]",
                    f"GRIB2 file from \033[38;5;202m{self.grib_source}\033[m and",
                    f"index file from \033[38;5;202m{self.idx_source}\033[m.",
                    f'{" ":150s}',
                )
            else:
                print(
                    f"💔 Did not find a GRIB2 or Index File for",
                    f"\033[32m{self.date:%Y-%b-%d %H:%M UTC} F{self.fxx:02d}\033[m",
                    f"{self.model.upper()}",
                    f'{" ":100s}',
                )

    def __repr__(self):
        """Representation in Notebook"""
        msg = (
            f"[{self.model.upper()}] model [{self.product}] product",
            f"run at \033[32m{self.date:%Y-%b-%d %H:%M UTC}",
            f"F{self.fxx:02d}\033[m",
        )
        return " ".join(msg)

    def __str__(self):
        """When class object is printed"""
        msg = [
            f"{self.model=}",
            f"{self.DETAILS=}",
            f"{self.DESCRIPTION=}",
            f"{self.product=}",
            f"{self.fxx=}",
            f"{self.date=}",
            f"{self.priority=}",
            f"{self.DETAILS=}",
            f"{self.SOURCES=}",
        ]
        return "\n".join(msg)

    def _validate(self):
        """Validate the Herbie class input arguments"""

        # Accept model alias
        if self.model.lower() == "alaska":
            self.model = "hrrrak"

        _models = {m for m in dir(models_template) if not m.startswith("__")}
        _products = set(self.PRODUCTS)

        assert self.date < datetime.utcnow(), "🔮 `date` cannot be in the future."
        assert self.model in _models, f"`model` must be one of {_models}"
        assert self.product in _products, f"`product` must be one of {_products}"

        if isinstance(self.priority, str):
            self.priority = [self.priority]

        if self.priority is not None:
            self.priority = [i.lower() for i in self.priority]

            # Don't look for data from NOMADS if requested date is earlier
            # than 14 days ago. NOMADS doesn't keep data that old,
            # (I think this is true of all models).
            if "nomads" in self.priority:
                expired = datetime.utcnow() - timedelta(days=14)
                expired = pd.to_datetime(f"{expired:%Y-%m-%d}")
                if self.date < expired:
                    self.priority.remove("nomads")

    def _ping_pando(self):
        """Pinging the Pando server before downloading can prevent a bad handshake."""
        try:
            requests.head("https://pando-rgw01.chpc.utah.edu/")
        except:
            print("🤝🏻⛔ Bad handshake with pando? Am I able to move on?")
            pass

    def _check_grib(self, url):
        """Check that the GRIB2 URL exist and is of useful length."""
        head = requests.head(url)
        check_exists = head.ok
        if check_exists:
            check_content = int(head.raw.info()["Content-Length"]) > 1_000_000
            return check_exists and check_content
        else:
            return False

    def _check_idx(self, url):
        """Check if an index file exist for the GRIB2 URL."""
        if not url.endswith(".idx"):
            url += ".idx"
        return requests.head(url).ok

    @property
    def get_remoteFileName(self, source=None):
        """Predict Remote File Name"""
        if source is None:
            source = list(self.SOURCES)[0]
        return self.SOURCES[source].split("/")[-1]

    @property
    def get_localFileName(self):
        """Predict Local File Name"""
        return self.LOCALFILE

    def get_localFilePath(self, searchString=None):
        """Get path to local file"""
        outFile = (
            self.save_dir.expand()
            / self.model
            / f"{self.date:%Y%m%d}"
            / self.get_localFileName
        )
        if searchString is not None:
            # Reassign the index DataFrame with the requested searchString
            self.idx_df = self.read_idx(searchString)

            # Get a list of all GRIB message numbers. We will use this
            # in the output file name as a unique identifier.
            all_grib_msg = "-".join([f"{i:g}" for i in self.idx_df.index])

            # Append the filename to distinguish it from the full file.
            outFile = outFile.with_suffix(f".grib2.subset_{all_grib_msg}")

        return outFile

    def read_idx(self, searchString=None):
        """
        Inspect the GRIB2 file contents by reading the index file.

        Parameters
        ----------
        searchString : str
            Filter dataframe by a searchString regular expression.
            Searches for strings in the index file lines, specifically
            the variable, level, and forecast_time columns.
            Execute ``_searchString_help()`` for examples of a good
            searchString.

            .. include:: ../user_guide/searchString.rst

        Returns
        -------
        A Pandas DataFrame of the index file.
        """
        assert self.idx is not None, f"No index file found for {self.grib}."

        # Sometimes idx end in ':', other times it doesn't (in some Pando files).
        # https://pando-rgw01.chpc.utah.edu/hrrr/sfc/20180101/hrrr.t00z.wrfsfcf00.grib2.idx
        # https://noaa-hrrr-bdp-pds.s3.amazonaws.com/hrrr.20210101/conus/hrrr.t00z.wrfsfcf00.grib2.idx
        # Sometimes idx has more than the standard messages
        # https://noaa-nbm-grib2-pds.s3.amazonaws.com/blend.20210711/13/core/blend.t13z.core.f001.co.grib2.idx

        r = requests.get(self.idx)
        assert r.ok, f"Index file does not exist: {self.idx}"
        df = pd.read_csv(
            self.idx,
            sep=":",
            names=[
                "grib_message",
                "start_byte",
                "reference_time",
                "variable",
                "level",
                "forecast_time",
                "?",
                "??",
                "???",
            ],
        )

        # Format the DataFrame
        df["grib_message"] = df["grib_message"].astype(float)
        # ^ float because RAP idx files have some decimal grib message numbers
        df["reference_time"] = pd.to_datetime(df.reference_time, format="d=%Y%m%d%H")
        df["valid_time"] = df["reference_time"] + pd.to_timedelta(f"{self.fxx}H")
        df["start_byte"] = df["start_byte"].astype(int)
        df["end_byte"] = df["start_byte"].shift(-1, fill_value="")
        df["range"] = df.start_byte.astype(str) + "-" + df.end_byte.astype(str)
        df = df.set_index("grib_message")
        df = df.reindex(
            columns=[
                "start_byte",
                "end_byte",
                "range",
                "reference_time",
                "valid_time",
                "variable",
                "level",
                "forecast_time",
                "?",
                "??",
                "???",
            ]
        )

        df = df.dropna(how="all", axis=1)
        df = df.fillna("")

        df.attrs = dict(
            url=self.idx,
            source=self.idx_source,
            description="Inventory index (.idx) file for the GRIB2 file.",
            model=self.model,
            product=self.product,
            lead_time=self.fxx,
            datetime=self.date,
        )

        # Filter DataFrame by searchString
        if searchString not in [None, ":"]:
            columns_to_search = df.loc[:, "variable":].apply(
                lambda x: ":".join(x).rstrip(":"), axis=1
            )
            logic = columns_to_search.str.contains(searchString)
            if logic.sum() == 0:
                print(
                    f"No GRIB messages found. There might be something wrong with {searchString=}"
                )
                print(_searchString_help(searchString))
            df = df.loc[logic]
        return df

    def download(
        self,
        searchString=None,
        *,
        source=None,
        save_dir=None,
        overwrite=None,
        verbose=True,
        errors="warn",
    ):
        """
        Download file from source.

        Subsetting by variable follows the same principles described here:
        https://www.cpc.ncep.noaa.gov/products/wesley/fast_downloading_grib.html

        Parameters
        ----------
        searchString : str
            If None, download the full file. Else, use regex to subset
            the file by specific variables and levels.
            .. include:: ../user_guide/searchString.rst
        source : {'nomads', 'aws', 'google', 'azure', 'pando', 'pando2'}
            If None, download GRIB2 file from self.grib2 which is
            the first location the GRIB2 file was found from the
            priority lists when this class was initialized. Else, you
            may specify the source to force downloading it from a
            different location.
        save_dir : str or pathlib.Path
            Location to save the model output files.
            If None, uses the default or path specified in __init__.
            Else, changes the path files are saved.
        overwrite : bool
            If True, overwrite existing files. Default will skip
            downloading if the full file exists. Not applicable when
            when searchString is not None because file subsets might
            be unique.
        errors : {'warn', 'raise'}
            When an error occurs, send a warning or raise a value error.
        """

        def _reporthook(a, b, c):
            """
            Print download progress in megabytes.

            Parameters
            ----------
            a : Chunk number
            b : Maximum chunk size
            c : Total size of the download
            """
            chunk_progress = a * b / c * 100
            total_size_MB = c / 1000000.0
            print(
                f"\r🚛💨  Download Progress: {chunk_progress:.2f}% of {total_size_MB:.1f} MB\r",
                end="",
            )

        def subset(searchString, outFile):
            """Download a subset specified by the regex searchString"""
            grib_source = self.grib
            if hasattr(grib_source, "as_posix") and grib_source.exists():
                # The GRIB source is local. Curl the local file
                # See https://stackoverflow.com/a/21023161/2383070
                grib_source = f"file://{str(self.grib)}"
            print(
                f'📇 Download subset: {self.__repr__()}{" ":60s}\n cURL from {grib_source}'
            )

            # Download subsets of the file by byte range with cURL.
            for i, (grbmsg, row) in enumerate(self.idx_df.iterrows()):
                print(
                    f"{i+1:>4g}: GRIB_message={grbmsg:<3g} \033[34m{':'.join(row.values[5:]).rstrip(':')}\033[m"
                )
                if i == 0:
                    # If we are working on the first item, overwrite the existing file...
                    curl = f"curl -s --range {row.range} {grib_source} > {outFile}"
                else:
                    # ...all other messages are appended to the subset file.
                    curl = f"curl -s --range {row.range} {grib_source} >> {outFile}"
                os.system(curl)

            self.local_grib_subset = outFile

        # If the file exists in the localPath and we don't want to
        # overwrite, then we don't need to download it.
        outFile = self.get_localFilePath(searchString=searchString)

        # This overrides the overwrite specified in __init__
        if overwrite is not None:
            self.overwrite = overwrite

        if outFile.exists() and not self.overwrite:
            if verbose:
                print(f"🌉 Already have local copy --> {outFile}")
            if searchString in [None, ":"]:
                self.local_grib = outFile
            else:
                self.local_grib_subset = outFile
            return

        # Attach the index file to the object (how much overhead is this?)
        if self.idx is not None:
            self.idx_df = self.read_idx(searchString)

        # This overrides the save_dir specified in __init__
        if save_dir is not None:
            self.save_dir = Path(save_dir).expand()

        if not hasattr(Path(self.save_dir).expand(), "exists"):
            self.save_dir = Path(self.save_dir).expand()

        # Check that data exists
        if self.grib is None:
            msg = f"🦨 GRIB2 file not found: {self.model=} {self.date=} {self.fxx=}"
            if errors == "warn":
                warnings.warn(msg)
                return  # Can't download anything without a GRIB file URL.
            elif errors == "raise":
                raise ValueError(msg)
        if self.idx is None and searchString is not None:
            msg = f"🦨 Index file not found; cannot download subset: {self.model=} {self.date=} {self.fxx=}"
            if errors == "warn":
                warnings.warn(
                    msg + " I will download the full file because I cannot subset."
                )
            elif errors == "raise":
                raise ValueError(msg)

        if source is not None:
            # Force download from a specified source and not from first in priority
            self.grib = self.SOURCES[source]

        # Create directory if it doesn't exist
        if not outFile.parent.is_dir():
            outFile.parent.mkdir(parents=True, exist_ok=True)
            print(f"👨🏻‍🏭 Created directory: [{outFile.parent}]")

        if searchString in [None, ":"] or self.idx is None:
            # Download the full file from remote source
            urllib.request.urlretrieve(self.grib, outFile, _reporthook)
            if verbose:
                print(
                    f"✅ Success! Downloaded {self.model.upper()} from \033[38;5;202m{self.grib_source:20s}\033[m\n\tsrc: {self.grib}\n\tdst: {outFile}"
                )
            self.local_grib = outFile
        else:
            # Download a subset of the file
            subset(searchString, outFile)

    def xarray(
        self, searchString, backend_kwargs={}, remove_grib=True, **download_kwargs
    ):
        """
        Open GRIB2 data as xarray DataSet

        Parameters
        ----------
        searchString : str
            Variables to read into xarray Dataset
        remove_grib : bool
            If True, grib file will be removed ONLY IF it didn't exist
            before we downloaded it.
        """

        download_kwargs = {**dict(overwrite=False), **download_kwargs}

        # Download file if local file does not exists
        local_file = self.get_localFilePath(searchString=searchString)

        # Only remove grib if it didn't exists before we download it
        remove_grib = not local_file.exists() and remove_grib

        if not local_file.exists() or download_kwargs["overwrite"]:
            self.download(searchString=searchString, **download_kwargs)

        # Backend kwargs for cfgrib
        backend_kwargs.setdefault("indexpath", "")
        backend_kwargs.setdefault(
            "read_keys", ["parameterName", "parameterUnits", "stepRange"]
        )
        backend_kwargs.setdefault("errors", "raise")

        # Use cfgrib.open_datasets, just in case there are multiple "hypercubes"
        # for what we requested.
        Hxr = cfgrib.open_datasets(
            self.get_localFilePath(searchString=searchString),
            backend_kwargs=backend_kwargs,
        )

        for h in Hxr:
            h.attrs["model"] = self.model
            h.attrs["product"] = self.product
            h.attrs["description"] = self.DESCRIPTION
            h.attrs["remote_grib"] = self.grib
            h.attrs["local_grib"] = self.get_localFilePath(searchString=searchString)

        if remove_grib:
            # Load the data to memory before removing the file
            Hxr = [ds.load() for ds in Hxr]
            # Only remove grib if it didn't exists before
            local_file.unlink()  # Removes file

        if len(Hxr) == 1:
            return Hxr[0]
        else:
            return Hxr



## Added by Brian Blaylock
## July 27, 2021

class nbm:
    def template(self):
        self.DESCRIPTION = 'National Blend of Models'
        self.DETAILS = {
            'NOMADS product description': 'https://www.nco.ncep.noaa.gov/pmb/products/blend/',
            'AWS Registry' : 'https://registry.opendata.aws/noaa-nbm/',
            'NWS National Blend of Models Home' : 'https://www.weather.gov/mdl/nbm_home',
            'MDL NBM Page': 'https://vlab.noaa.gov/web/mdl/nbm',
        }
        self.PRODUCTS = {
            'ak': 'Alaska; 13-km resolution',
            'co': 'CONUS 13-km resolution',
            'gu': 'Guam 13-km resolution',
            'hi': 'Hawaii 13-km resolution',
            'pr': 'Puerto Rico 13-km resolution',
        }
        self.SOURCES = {
            'nomads': f'https://nomads.ncep.noaa.gov/pub/data/nccf/com/blend/prod/blend.{self.date:%Y%m%d/%H}/core/blend.t{self.date:%H}z.core.f{self.fxx:03d}.{self.product}.grib2',
            'aws': f'https://noaa-nbm-grib2-pds.s3.amazonaws.com/blend.{self.date:%Y%m%d/%H}/core/blend.t{self.date:%H}z.core.f{self.fxx:03d}.{self.product}.grib2',
        }
        self.LOCALFILE = f"{self.get_remoteFileName}"


## Added by Brian Blaylock
## July 28, 2021

class navgem:
    def template(self):
        self.DESCRIPTION = 'Navy Global Environment Model'
        self.DETAILS = {
            'NRL description': 'https://www.nrlmry.navy.mil/metoc/nogaps/navgem.html',
        }
        self.PRODUCTS = {
            'none': '',
        }
        self.SOURCES = {
            'nomads': f'https://nomads.ncep.noaa.gov/pub/data/nccf/com/fnmoc/prod/navgem.{self.date:%Y%m%d}/navgem_{self.date:%Y%m%d%H}f{self.fxx:03d}.grib2',
        }
        self.LOCALFILE = f"{self.get_remoteFileName}"


## Added by Brian Blaylock
## July 26, 2021

class gfs_wave:
    def template(self):
        self.DESCRIPTION = 'Global Forecast System - Wave Products'
        self.DETAILS = {
            'nomads product description': 'https://www.nco.ncep.noaa.gov/pmb/products/gfs/#GFSwave',
        }
        self.PRODUCTS = {
            'arctic.9km': 'Arctic; 9-km resolution',
            'atlocn.0p16': 'North Atlantic 0.16 deg resolution',
            'epacif.0p16': 'Eastern Pacific; .16 deg resolution',
            'global.0p16': 'Global; 0.16 deg resolution',
            'global.0p25': 'Global; 0.25 deg resolution',
            'gsouth.0p25': 'Global South; 0.25 deg resolution',
            'wcoast.0p16': 'West Coast; 0.16 deg resolution',

        }
        self.SOURCES = {
            'nomads': f'https://nomads.ncep.noaa.gov/pub/data/nccf/com/gfs/prod/gfs.{self.date:%Y%m%d/%H}/wave/gridded/gfswave.t{self.date:%H}z.{self.product}.f{self.fxx:03d}.grib2',
            'aws': f'https://noaa-gfs-bdp-pds.s3.amazonaws.com/gfs.{self.date:%Y%m%d/%H}/wave/gridded/gfswave.t{self.date:%H}z.{self.product}.f{self.fxx:03d}.grib2',
            'google': f'https://storage.googleapis.com/global-forecast-system/gfs.{self.date:%Y%m%d/%H}/wave/gridded/gfswave.t{self.date:%H}z.{self.product}.f{self.fxx:03d}.grib2',
            'azure': f'https://noaahrrr.blob.core.windows.net/gfs/gfs.{self.date:%Y%m%d/%H}/wave/gridded/gfswave.t{self.date:%H}z.{self.product}.f{self.fxx:03d}.grib2',
        }
        self.LOCALFILE = f"{self.get_remoteFileName}"


## Brian Blaylock
## April 28, 2021

import warnings
import configparser
from pathlib import Path

warnings.warn(
      "The hrrrb API is deprecated. Use the new Herbie API instead.",
      DeprecationWarning
   )

########################################################################
# Load custom xarray accessors
try:
    import hrrrb.accessors
except:
    warnings.warn("HRRR-B's xarray accessors could not be imported.")
    pass

########################################################################
# Configure HRRR-B
# Configuration file is save in `~/config/hrrrb/config.cfg`
# `_default_save_dir` is the default path to save GRIB2 files.

config = configparser.ConfigParser()
_config_path = Path('~').expanduser() / '.config' / 'hrrrb' / 'config.cfg'

user_home_default = str(Path('~').expanduser() / 'data')

if not _config_path.exists():
    _config_path.parent.mkdir(parents=True)
    _config_path.touch()
    config.read(_config_path)
    config.add_section('download')
    config.set('download', 'default_save_dir', user_home_default)
    with open(_config_path, 'w') as configfile:
        config.write(configfile)
    print(f'⚙ Created config file [{_config_path}]',
          f'with default download directory set as [{user_home_default}]')

config.read(_config_path)

try:
    _default_save_dir = Path(config.get('download', 'default_save_dir'))
except:
    print(f'🦁🐯🐻 oh my! {_config_path} looks weird,',
          f'but I will add a new section')
    config.add_section('download')
    config.set('download', 'default_save_dir', user_home_default)
    with open(_config_path, 'w') as configfile:
        config.write(configfile)
    _default_save_dir = Path(config.get('download', 'default_save_dir'))


## Brian Blaylock
## April 23, 2021

"""
==================================
Herbie Extension: xarray accessors
==================================

Extend the xarray capabilities with a custom accessor.
http://xarray.pydata.org/en/stable/internals.html#extending-xarray

To use the herbie xarray accessor, do this...

.. code-block:: python

    H = Herbie('2021-01-01', model='hrrr')
    ds = H.xarray('TMP:2 m')
    ds.herbie.crs
    ds.herbie.plot

"""
import warnings

import numpy as np
import pandas as pd
import xarray as xr
import matplotlib.pyplot as plt
import cartopy.crs as ccrs


# From Carpenter_Workshop:
# https://github.com/blaylockbk/Carpenter_Workshop
from toolbox.cartopy_tools import common_features, pc
from paint.standard2 import cm_tmp, cm_dpt, cm_rh, cm_wind, cm_pcp
from paint.radar2 import cm_reflectivity
from paint.radar import cm_reflectivity


@xr.register_dataset_accessor("herbie")
class HerbieAccessor:
    def __init__(self, xarray_obj):
        self._obj = xarray_obj
        self._center = None

    @property
    def center(self):
        """Return the geographic center point of this dataset."""
        if self._center is None:
            # we can use a cache on our accessor objects, because accessors
            # themselves are cached on instances that access them.
            lon = self._obj.latitude
            lat = self._obj.longitude
            self._center = (float(lon.mean()), float(lat.mean()))
        return self._center

    @property
    def crs(self):
        """
        Cartopy coordinate reference system (crs) from a cfgrib Dataset.

        Projection information is from the grib2 message for each variable.

        Parameters
        ----------
        ds : xarray.Dataset
            An xarray.Dataset from a GRIB2 file opened by the cfgrib engine.
        """

        ds = self._obj

        # Get projection from the attributes of HRRR xarray.Dataset
        # CF 1.8 map projection information for the HRRR model
        # http://cfconventions.org/Data/cf-conventions/cf-conventions-1.8/cf-conventions.html#_lambert_conformal

        # Projection info is in the variable attributes. The first variable will do.
        attrs = ds[list(ds)[0]].attrs

        if attrs["GRIB_gridType"] == "lambert":
            lc_kwargs = dict(
                globe=ccrs.Globe(ellipse="sphere"),
                central_latitude=attrs["GRIB_LaDInDegrees"],
                central_longitude=attrs["GRIB_LoVInDegrees"],
                standard_parallels=(
                    attrs["GRIB_Latin1InDegrees"],
                    attrs["GRIB_Latin2InDegrees"],
                ),
            )
            lc = ccrs.LambertConformal(**lc_kwargs)
            return lc
        elif attrs["GRIB_gridType"] == "polar_stereographic":
            npc_kwargs = {}
            if ds.model == "hrrrak":
                # See https://rapidrefresh.noaa.gov/hrrr/ALASKA/static/
                npc_kwargs["central_longitude"] = -135
            npc = ccrs.NorthPolarStereo(**npc_kwargs)
            return npc
        elif attrs["GRIB_gridType"] == "regular_ll":
            pc_kwargs = {}
            if attrs["GRIB_longitudeOfFirstGridPointInDegrees"] == 0:
                pc_kwargs["central_longitude"] = 180
            pc = ccrs.PlateCarree(**pc_kwargs)
            return pc
        else:
            warnings.warn(
                f'GRIB_gridType [{attrs["GRIB_gridType"]}] is not recognized.'
            )
            return None

    def plot(self, ax=None, common_features_kw={}, **kwargs):
        """Plot data on a map."""
        ds = self._obj

        _level_units = dict(
            adiabaticCondensation="adiabatic condensation",
            atmosphere="atmosphere",
            atmosphereSingleLayer="atmosphere single layer",
            boundaryLayerCloudLayer="boundary layer cloud layer",
            cloudBase="cloud base",
            cloudCeiling="cloud ceiling",
            cloudTop="cloud top",
            depthBelowLand="m",
            equilibrium="equilibrium",
            heightAboveGround="m",
            heightAboveGroundLayer="m",
            highCloudLayer="high cloud layer",
            highestTroposphericFreezing="highest tropospheric freezing",
            isobaricInhPa="hPa",
            isobaricLayer="hPa",
            isothermZero="0 C",
            isothermal="K",
            level="m",
            lowCloudLayer="low cloud layer",
            meanSea="MSLP",
            middleCloudLayer="middle cloud layer",
            nominalTop="nominal top",
            pressureFromGroundLayer="Pa",
            sigma="sigma",
            sigmaLayer="sigmaLayer",
            surface="surface",
        )

        for var in ds.data_vars:
            print("cfgrib variable:", var)
            print("GRIB_cfName", ds[var].GRIB_cfName)
            print("GRIB_cfVarName", ds[var].GRIB_cfVarName)
            print("GRIB_name", ds[var].GRIB_name)
            print("GRIB_units", ds[var].GRIB_units)
            print("GRIB_typeOfLevel", ds[var].GRIB_typeOfLevel)
            print()

            ds[var].attrs["units"] = (
                ds[var]
                .attrs["units"]
                .replace("**-1", "$^{-1}$")
                .replace("**-2", "$^{-2}$")
            )

            dpi = common_features_kw.pop("dpi", 150)
            figsize = common_features_kw.pop("figsize", [10, 5])
            fig, ax = plt.subplots(
                1,
                1,
                subplot_kw=dict(projection=ds.herbie.crs),
                dpi=dpi,
                figsize=figsize,
            )

            default = dict(
                scale="50m", ax=ax, crs=ds.herbie.crs, STATES=True, BORDERS=True
            )
            common_features_kw = {**default, **common_features_kw}
            common_features(**common_features_kw)

            title = ""
            kwargs = {}
            cbar_kwargs = dict(pad=0.01)

            if ds[var].GRIB_cfVarName in ["d2m", "dpt"]:
                ds[var].attrs["GRIB_cfName"] = "dew_point_temperature"

            ## Wind
            wind_pair = {"u10": "v10", "u80": "v80", "u": "v"}

            if ds[var].GRIB_cfName == "air_temperature":
                kwargs = {**cm_tmp().cmap_kwargs, **kwargs}
                cbar_kwargs = {**cm_tmp().cbar_kwargs, **cbar_kwargs}
                if ds[var].GRIB_units == "K":
                    ds[var] -= 273.15
                    ds[var].attrs["GRIB_units"] = "C"
                    ds[var].attrs["units"] = "C"

            elif ds[var].GRIB_cfName == "dew_point_temperature":
                kwargs = {**cm_dpt().cmap_kwargs, **kwargs}
                cbar_kwargs = {**cm_dpt().cbar_kwargs, **cbar_kwargs}
                if ds[var].GRIB_units == "K":
                    ds[var] -= 273.15
                    ds[var].attrs["GRIB_units"] = "C"
                    ds[var].attrs["units"] = "C"

            elif ds[var].GRIB_name == "Total Precipitation":
                title = "-".join(
                    [f"F{int(i):02d}" for i in ds[var].GRIB_stepRange.split("-")]
                )
                ds[var] = ds[var].where(ds[var] != 0)
                kwargs = {**cm_pcp().cmap_kwargs, **kwargs}
                cbar_kwargs = {**cm_pcp().cbar_kwargs, **cbar_kwargs}

            elif ds[var].GRIB_name == "Maximum/Composite radar reflectivity":
                ds[var] = ds[var].where(ds[var] >= 0)
                cbar_kwargs = {**cm_reflectivity().cbar_kwargs, **cbar_kwargs}
                kwargs = {**cm_reflectivity().cmap_kwargs, **kwargs}

            elif ds[var].GRIB_cfName == "relative_humidity":
                cbar_kwargs = {**cm_rh().cbar_kwargs, **cbar_kwargs}
                kwargs = {**cm_rh().cmap_kwargs, **kwargs}

            elif "wind" in ds[var].GRIB_cfName or "wind" in ds[var].GRIB_name:
                cbar_kwargs = {**cm_wind().cbar_kwargs, **cbar_kwargs}
                kwargs = {**cm_wind().cmap_kwargs, **kwargs}
                if ds[var].GRIB_cfName == "eastward_wind":
                    cbar_kwargs["label"] = "U " + cbar_kwargs["label"]
                elif ds[var].GRIB_cfName == "northward_wind":
                    cbar_kwargs["label"] = "V " + cbar_kwargs["label"]
            else:
                cbar_kwargs = {
                    **dict(
                        label=f"{ds[var].GRIB_parameterName.strip().title()} ({ds[var].units})"
                    ),
                    **cbar_kwargs,
                }

            p = ax.pcolormesh(
                ds.longitude, ds.latitude, ds[var], transform=pc, **kwargs
            )
            plt.colorbar(p, ax=ax, **cbar_kwargs)

            VALID = pd.to_datetime(ds.valid_time.data).strftime("%H:%M UTC %d %b %Y")
            RUN = pd.to_datetime(ds.time.data).strftime("%H:%M UTC %d %b %Y")
            FXX = f"F{pd.to_timedelta(ds.step.data).total_seconds()/3600:02.0f}"

            level_type = ds[var].GRIB_typeOfLevel
            if level_type in _level_units:
                level_units = _level_units[level_type]
            else:
                level_units = "unknown"

            if level_units.lower() in ["surface", "atmosphere"]:
                level = f"{title} {level_units}"
            else:
                level = f"{ds[var][level_type].data:g} {level_units}"

            ax.set_title(
                f"Run: {RUN} {FXX}",
                loc="left",
                fontfamily="monospace",
                fontsize="x-small",
            )
            ax.set_title(
                f"{ds.model.upper()} {level}", loc="center", fontweight="semibold"
            )
            ax.set_title(
                f"Valid: {VALID}",
                loc="right",
                fontfamily="monospace",
                fontsize="x-small",
            )

            # Set extent (could do this more efficiently by storing the data elsewhere)
            new = ds.herbie.crs.transform_points(
                pc, ds.longitude.data, ds.latitude.data
            )
            LONS = new[:, :, 0]
            LATS = new[:, :, 1]
            ax.set_extent(
                [LONS.min(), LONS.max(), LATS.min(), LATS.max()], crs=ds.herbie.crs
            )

        return ax



## Added by Brian Blaylock
## July 26, 2021

class gfs:
    def template(self):
        self.DESCRIPTION = 'Global Forecast System'
        self.DETAILS = {
            'nomads product description': 'https://www.nco.ncep.noaa.gov/pmb/products/gfs',
            'google cloud platform': 'https://console.cloud.google.com/marketplace/product/noaa-public/gfs?q=search&referrer=search&project=python-232920',
            'azure document':'https://github.com/microsoft/AIforEarthDatasets#noaa-global-forecast-system-gfs',
        }
        self.PRODUCTS = {
            'pgrb2.0p25': 'common fields, 0.25 degree resolution',
            'pgrb2.0p50': 'common fields, 0.50 degree resolution',
            'pgrb2.1p00': 'common fields, 1.00 degree resolution',
            'pgrb2b.0p25': 'uncommon fields, 0.25 degree resolution',
            'pgrb2b.0p50': 'uncommon fields, 0.50 degree resolution',
            'pgrb2b.1p00': 'uncommon fields, 1.00 degree resolution',
            'pgrb2full.0p50': 'combined grids of 0.50 resolution',
        }
        self.SOURCES = {
            'aws': f'https://noaa-gfs-bdp-pds.s3.amazonaws.com/gfs.{self.date:%Y%m%d/%H}/atmos/gfs.t{self.date:%H}z.{self.product}.f{self.fxx:03d}',
            'aws-old': f'https://noaa-gfs-bdp-pds.s3.amazonaws.com/gfs.{self.date:%Y%m%d/%H}/gfs.t{self.date:%H}z.{self.product}.f{self.fxx:03d}',
            'nomads': f'https://nomads.ncep.noaa.gov/pub/data/nccf/com/gfs/prod/gfs.{self.date:%Y%m%d/%H}/atmos/gfs.t{self.date:%H}z.{self.product}.f{self.fxx:03d}',
            'google': f'https://storage.googleapis.com/global-forecast-system/gfs.{self.date:%Y%m%d/%H}/atmos/gfs.t{self.date:%H}z.{self.product}.f{self.fxx:03d}',
            'azure': f'https://noaagfs.blob.core.windows.net/gfs/gfs.{self.date:%Y%m%d/%H}/atmos/gfs.t{self.date:%H}z.{self.product}.f{self.fxx:03d}',
        }
        self.LOCALFILE = f"{self.get_remoteFileName}"


## Added by Brian Blaylock
## July 26, 2021

# NOTE: The Rapid Refresh Forecast System is rapidly changing

class rrfs:
    def template(self):
        self.DESCRIPTION = 'Rapid Refresh Forecast System (RRFS) Ensemble'
        self.DETAILS = {
            'aws product description': 'https://registry.opendata.aws/noaa-rrfs/',
        }
        self.PRODUCTS = {
            # Below are ensemble products found in ensprod/
            'mean' : 'ensemble mean',
            'avrg' : 'ensemble products: ???',
            'eas' : 'ensemble products: ???',
            'ffri' : 'ensemble products: ???',
            'lpmm' : 'ensemble products: ???',
            'pmmn' : 'ensemble products: ???',
            'prob' : 'ensemble products: ???',
            # Below are member products found in mem##/
            'testbed.conus': 'surface grids (one for each member)',
            'na': 'native grids (one for each member)',
        }
        self.SOURCES = {
            'aws' : f'https://noaa-rrfs-pds.s3.amazonaws.com/rrfs.{self.date:%Y%m%d/%H}/ensprod/rrfsce.t{self.date:%H}z.conus.{self.product}.f{self.fxx:02d}.grib2',
            'aws-mem' : f'https://noaa-rrfs-pds.s3.amazonaws.com/rrfs.{self.date:%Y%m%d/%H}/mem{self.member:02d}/rrfs.t{self.date:%H}z.mem{self.member:02d}.{self.product}f{self.fxx:03d}.grib2',
        }
        self.LOCALFILE = f"mem{self.member:02d}/{self.get_remoteFileName}"